#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
# TODO: need better coverage for different tile types

from io import StringIO
import os
import os.path
import re
import unittest
import tempfile

import prjxray
import utils.fasm2frames as fasm2frames
from prjxray.util import OpenSafeFile

from textx.exceptions import TextXSyntaxError


def frm2bits(txt):
    '''
    Convert output .frm file text to set of (frame addr, word #, bit index) tuples
    '''
    bits_out = set()
    for l in txt.split('\n'):
        l = l.strip()
        if not l:
            continue
        # 0x00020500 0x00000000,0x00000000,0x00000000,...
        addr, words = l.split(' ')
        addr = int(addr, 0)
        words = words.split(',')
        assert (101 == len(words))
        for wordi, word in enumerate(words):
            word = int(word, 0)
            for biti in range(32):
                val = word & (1 << biti)
                if val:
                    bits_out.add((addr, wordi, biti))
    return bits_out


def bitread2bits(txt):
    '''
    Convert .bits text file (ie bitread output) to set of (frame addr, word #, bit index) tuples
    '''
    bits_ref = set()
    for l in txt.split('\n'):
        l = l.strip()
        if not l:
            continue
        # bit_0002050b_004_14
        m = re.match(r'bit_(.{8})_(.{3})_(.{2})', l)
        addr = int(m.group(1), 16)
        word = int(m.group(2), 10)
        bit = int(m.group(3), 10)
        bits_ref.add((addr, word, bit))
    return bits_ref


class TestStringMethods(unittest.TestCase):
    def filename_test_data(self, fname):
        return os.path.join(os.path.dirname(__file__), 'test_data', fname)

    def get_test_data(self, fname):
        with OpenSafeFile(self.filename_test_data(fname)) as f:
            return f.read()

    def fasm2frames(self, fin_data, **kw):
        with tempfile.NamedTemporaryFile(suffix='.fasm') as fin:
            fin.write(fin_data.encode('utf-8'))
            fin.flush()

            fout = StringIO()
            fasm2frames.run(
                self.filename_test_data('db'), "xc7a200tffg1156-1", fin.name,
                fout, **kw)

            return fout.getvalue()

    def test_lut(self):
        '''Simple smoke test on just the LUTs'''
        self.fasm2frames(self.get_test_data('lut.fasm'))

    def bitread_frm_equals(self, frm_fn, bitread_fn):
        fout = self.fasm2frames(self.get_test_data(frm_fn))

        # Build a list of output used bits
        bits_out = frm2bits(fout)

        # Build a list of reference used bits
        bits_ref = bitread2bits(self.get_test_data(bitread_fn))

        # Now check for equivilence vs reference design
        self.assertEqual(len(bits_ref), len(bits_out))
        self.assertEqual(bits_ref, bits_out)

    def test_lut_int(self):
        self.bitread_frm_equals('lut_int.fasm', 'lut_int/design.bits')

    def test_ff_int(self):
        self.bitread_frm_equals('ff_int.fasm', 'ff_int/design.bits')

    @unittest.skip
    def test_ff_int_op1(self):
        '''Omitted key set to '''
        self.bitread_frm_equals('ff_int_op1.fasm', 'ff_int/design.bits')

    # Same check as above, but isolated test case
    def test_opkey_01_default(self):
        '''Optional key with binary omitted value should produce valid result'''
        self.fasm2frames("CLBLM_L_X10Y102.SLICEM_X0.SRUSEDMUX")

    @unittest.skip
    def test_opkey_01_1(self):
        self.fasm2frames("CLBLM_L_X10Y102.SLICEM_X0.SRUSEDMUX 1")

    @unittest.skip
    def test_opkey_enum(self):
        '''Optional key with enumerated value should produce syntax error'''
        try:
            # CLBLM_L.SLICEM_X0.AMUXFF.O6 !30_06 !30_07 !30_08 30_11
            self.fasm2frames("CLBLM_L_X10Y102.SLICEM_X0.AFFMUX.O6")
            self.fail("Expected syntax error")
        except TextXSyntaxError:
            pass

    def test_ff_int_0s(self):
        '''Explicit 0 entries'''
        self.bitread_frm_equals('ff_int_0s.fasm', 'ff_int/design.bits')

    def test_badkey(self):
        '''Bad key should throw syntax error'''
        try:
            self.fasm2frames("CLBLM_L_X10Y102.SLICEM_X0.SRUSEDMUX 2")
            self.fail("Expected syntax error")
        except TextXSyntaxError:
            pass

    @unittest.skip
    def test_dupkey(self):
        '''Duplicate key should throw syntax error'''
        try:
            self.fasm2frames(
                """\
CLBLM_L_X10Y102.SLICEM_X0.SRUSEDMUX 0
CLBLM_L_X10Y102.SLICEM_X0.SRUSEDMUX 1
""")
            self.fail("Expected syntax error")
        except TextXSyntaxError:
            pass

    @unittest.skip
    def test_sparse(self):
        '''Verify sparse equivalent to normal encoding'''
        frm_fn = 'lut_int.fasm'

        fout_sparse_txt = self.fasm2frames(
            self.get_test_data(frm_fn), sparse=True)
        bits_sparse = frm2bits(fout_sparse_txt)

        fout_full_txt = self.fasm2frames(
            self.get_test_data(frm_fn), sparse=False)
        bits_full = frm2bits(fout_full_txt)

        # Now check for equivilence vs reference design
        self.assertEquals(len(bits_sparse), len(bits_full))
        self.assertEquals(bits_sparse, bits_full)

        # Verify the full ROI is way bigger description
        # It will still be decent size though since even sparse occupies all columns in that area
        self.assertGreaterEqual(len(fout_full_txt), len(fout_sparse_txt) * 4)

    def test_stepdown_1(self):
        self.bitread_frm_equals(
            'iob/liob_stepdown.fasm', 'iob/liob_stepdown.bits')

    def test_stepdown_2(self):
        self.bitread_frm_equals(
            'iob/riob_stepdown.fasm', 'iob/riob_stepdown.bits')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="prjxray",
    version="0.0.1",
    author="SymbiFlow Authors",
    author_email="symbiflow@lists.librecores.org",
    description="Project X-Ray libraries",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/SymbiFlow/prjxray",
    packages=['prjxray'],
    install_requires=[
        'fasm',
        'intervaltree',
        'numpy',
        # FIXME: remove dependency once https://github.com/SymbiFlow/prjxray/issues/1624
        #        is fixed
        'pyjson5',
        'pyyaml',
        'simplejson',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: ISC License",
        "Operating System :: OS Independent",
    ],
    entry_points={
        'console_scripts': ['fasm2frames=utils.fasm2frames:main'],
    })



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

from prjxray.segmaker import Segmaker
from prjxray.db import Database
from prjxray.util import get_db_root, get_part
import re

REBUF_GCLK = re.compile('^CLK_BUFG_REBUF_R_CK_GCLK([0-9]+)_BOT$')

GCLKS = 32


def gclk_of_wire(wire):
    m = REBUF_GCLK.match(wire)
    assert m, wire
    return int(m.group(1))


class ClockColumn(object):
    def __init__(self, db_root, part):
        db = Database(db_root, part)
        grid = db.grid()

        tiles_in_gclk_columns = []
        self.gclk_columns = {}

        for tile in grid.tiles():
            gridinfo = grid.gridinfo_at_tilename(tile)

            if gridinfo.tile_type != 'CLK_BUFG_REBUF':
                continue

            loc = grid.loc_of_tilename(tile)

            tiles_in_gclk_columns.append((loc.grid_y, tile))

        _, self.tiles_in_gclk_columns = zip(
            *sorted(tiles_in_gclk_columns, key=lambda x: x[0]))

        # Initially all GCLK lines are idle.  GCLK lines only exist between
        #CLK_BUFG_REBUF tiles, hence len-1.
        for gclk in range(GCLKS):
            self.gclk_columns[gclk] = [
                False for _ in range(len(self.tiles_in_gclk_columns) - 1)
            ]

    def enable_rebuf(self, tile, wire):
        # Find which REBUF is being activated.
        rebuf_idx = self.tiles_in_gclk_columns.index(tile)
        assert rebuf_idx != -1, tile

        gclk = gclk_of_wire(wire)

        self.gclk_columns[gclk][rebuf_idx] = True
        self.gclk_columns[gclk][rebuf_idx - 1] = True

    def yield_rebuf_state(self):
        """ Yields tile_name, gclk, bool if active above tile, bool if active below tile """
        for idx, tile in enumerate(self.tiles_in_gclk_columns):
            for gclk in range(GCLKS):
                active_below = False
                active_above = False

                if idx > 0:
                    active_below = self.gclk_columns[gclk][idx - 1]

                if idx < len(self.gclk_columns[gclk]):
                    active_above = self.gclk_columns[gclk][idx]

                yield tile, gclk, active_below, active_above


def main():
    db_root = get_db_root()
    part = get_part()

    clock_column = ClockColumn(db_root, part)

    segmk = Segmaker("design.bits")

    print("Loading tags from design.txt.")

    with open("route.txt", "r") as f:
        for line in f:
            if 'CLK_BUFG_REBUF' not in line:
                continue

            parts = line.replace('{', '').replace('}', '').strip().replace(
                '\t', ' ').split(' ')
            dst = parts[0]
            pip = parts[3]

            tile_from_pip, pip = pip.split('/')

            if 'CLK_BUFG_REBUF' not in tile_from_pip:
                continue

            tile_type, pip = pip.split('.')
            assert tile_type == 'CLK_BUFG_REBUF'

            wire_a, wire_b = pip.split('<<->>')

            tile_from_wire, dst = dst.split('/')

            assert dst == wire_a

            if tile_from_wire == tile_from_pip:
                b_to_a = wire_a == dst
                a_to_b = not b_to_a
            else:
                b_to_a = wire_a != dst
                a_to_b = not b_to_a

            segmk.add_tile_tag(
                tile_from_pip, '{}.{}'.format(wire_a, wire_b), b_to_a)
            segmk.add_tile_tag(
                tile_from_pip, '{}.{}'.format(wire_b, wire_a), a_to_b)

            clock_column.enable_rebuf(tile_from_pip, wire_a)

    for tile, gclk, active_below, active_above in clock_column.yield_rebuf_state(
    ):
        segmk.add_tile_tag(
            tile, 'GCLK{}_ENABLE_ABOVE'.format(gclk), active_above)
        segmk.add_tile_tag(
            tile, 'GCLK{}_ENABLE_BELOW'.format(gclk), active_below)

    segmk.compile()
    segmk.write(allow_empty=True)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import sys, os, argparse
from sdf_timing import sdfparse

model_carry4 = {
    'type': 'CARRY4',
    'srcs': {'O5', '?X', '?X_LFF', '?X_LBOTH'},
    'out': {'CO0', 'CO1', 'CO2', 'CO3', 'O0', 'O1', 'O2', 'O3'},
    'mux': '?CY0',
    'pins': {
        'DI0': {
            'type': 'A'
        },
        'DI1': {
            'type': 'B'
        },
        'DI2': {
            'type': 'C'
        },
        'DI3': {
            'type': 'D'
        },
    },
}


def compute_delays(model, fin_name, fout_name):
    data = ''
    with open(fin_name, 'r') as f:
        data = f.read()
    sdf = sdfparse.parse(data)

    keys = sdf['cells'][model['type']].keys()
    if 'slicel'.upper() in keys:
        sl = 'L'
    elif 'slicem'.upper() in keys:
        sl = 'M'
    else:
        print("Unknown slice type!")
        return

    nsdf = dict()
    nsdf['header'] = sdf['header']
    nsdf['cells'] = dict()
    nsdf['cells']['ROUTING_BEL'] = dict()

    for p in model['pins']:
        pin = model['pins'][p]

        outs = dict()
        for o in model['out']:
            outs[o] = []

        res = []
        cells = sdf['cells']

        for src in model['srcs']:
            source = src.replace('?', pin['type'])

            _type = model['type'] + '_' + source

            if _type in cells.keys():
                cell = cells[_type]["SLICE" + sl.upper()]

                for o in model['out']:
                    iopath = 'iopath_' + p + '_' + o

                    if iopath in cell.keys():
                        delay = cell[iopath]['delay_paths']['slow']['max']
                        outs[o].append(delay)

        for src in outs:
            s = sorted(outs[src])
            for val in s:
                res.append(val - s[0])

        delay = round(max(res), 3)

        muxname = str(model['mux'].replace('?', pin['type']))
        rbel = nsdf['cells']['ROUTING_BEL']['SLICE' + sl.upper() + '/' +
                                            muxname] = dict()

        iname = 'interconnect_' + pin['type'].lower() + 'x_' + str(p).lower()

        rbel[iname] = dict()
        rbel[iname]['is_absolute'] = True
        rbel[iname]['to_pin_edge'] = None
        rbel[iname]['from_pin_edge'] = None
        rbel[iname]['from_pin'] = pin['type'].lower() + 'x'
        rbel[iname]['to_pin'] = str(p).lower()
        rbel[iname]['type'] = 'interconnect'
        rbel[iname]['is_timing_check'] = False
        rbel[iname]['is_timing_env'] = False

        paths = rbel[iname]['delay_paths'] = dict()

        paths['slow'] = dict()
        paths['slow']['min'] = delay
        paths['slow']['avg'] = None
        paths['slow']['max'] = delay

        paths['fast'] = dict()
        paths['fast']['min'] = delay
        paths['fast']['avg'] = None
        paths['fast']['max'] = delay

        # emit new sdfs
        with open(fout_name, 'w') as f:
            f.write(sdfparse.emit(nsdf))


def main(argv):
    parser = argparse.ArgumentParser(
        description='Tool for computing CARRY4 muxes delays')
    parser.add_argument(
        '--input', dest='inputfile', action='store', help='Input file')
    parser.add_argument(
        '--output', dest='outputfile', action='store', help='Output file')
    args = parser.parse_args(argv[1:])

    compute_delays(model_carry4, args.inputfile, args.outputfile)


if __name__ == "__main__":
    main(sys.argv)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

from collections import namedtuple
WireInGrid = namedtuple('WireInGrid', 'tile grid_x grid_y wire')
Connection = namedtuple('Connection', 'wire_a wire_b')


class Connections(object):
    def __init__(self, tilegrid, tileconn, tile_wires):
        self.grid = tilegrid
        self.tile_wires = tile_wires
        self.coord_to_tile = {}
        self.coord_to_tile_type = {}

        for tile, tile_info in self.grid.items():
            self.coord_to_tile[(tile_info['grid_x'],
                                tile_info['grid_y'])] = tile
            self.coord_to_tile_type[(tile_info['grid_x'],
                                     tile_info['grid_y'])] = tile_info['type']

            # Make sure we have tile type info for every tile in the grid.
            assert tile_info['type'] in self.tile_wires, (
                tile_info['type'], self.tile_wires.keys())

        self.potential_connections = {}

        for conn in tileconn:
            grid_deltas = conn['grid_deltas']
            tile_types = conn['tile_types']

            for pairs in conn['wire_pairs']:
                key = (tile_types[0], pairs[0])
                if key not in self.potential_connections:
                    self.potential_connections[key] = []
                self.potential_connections[key].append(
                    (grid_deltas, tile_types[1], pairs[1]))

    def all_possible_connections_from(self, wire_in_grid):
        tile_type = self.coord_to_tile_type[(
            wire_in_grid.grid_x, wire_in_grid.grid_y)]

        key = (tile_type, wire_in_grid.wire)

        if key not in self.potential_connections:
            return

        for relative_coord, target_tile_type, target_wire in (
                self.potential_connections[key]):
            rel_x, rel_y = relative_coord
            target_coord = (
                wire_in_grid.grid_x + rel_x, wire_in_grid.grid_y + rel_y)

            if target_coord in self.coord_to_tile_type:
                if self.coord_to_tile_type[target_coord] == target_tile_type:
                    yield Connection(
                        wire_in_grid,
                        WireInGrid(
                            tile=self.coord_to_tile[target_coord],
                            grid_x=target_coord[0],
                            grid_y=target_coord[1],
                            wire=target_wire))

    def get_connections(self):
        """ Yields Connection objects that represent all connections present in
    the grid based on tileconn """
        for tile, tile_info in self.grid.items():
            for wire in self.tile_wires[tile_info['type']]:
                wire_in_grid = WireInGrid(
                    tile=tile,
                    grid_x=tile_info['grid_x'],
                    grid_y=tile_info['grid_y'],
                    wire=wire)
                for potential_connection in self.all_possible_connections_from(
                        wire_in_grid):
                    yield potential_connection



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import os
import random
import json
import csv
from collections import defaultdict

try:
    random.seed(int(os.getenv("SEED"), 16))
except TypeError:
    pass

# =============================================================================


def load_iob_sites(file_name):
    """
    Loads IOB site dump from the given CSV file.
    """

    # Load the data
    with open(file_name, "r") as fp:
        data = [row for row in csv.DictReader(fp)]

    # Index IOB site data by clock regions
    iob_sites = defaultdict(lambda: [])
    for site_data in data:
        iob_sites[site_data["clock_region"]].append(site_data)

    return iob_sites


# =============================================================================

IOBUF_NOT_ALLOWED = [
    'HSTL_I',
    'HSTL_I_18',
    'SSTL18_I',
]

DIFF_MAP = {
    'SSTL135': 'DIFF_SSTL135',
    'SSTL15': 'DIFF_SSTL15',
}

VREF_ALLOWED = [
    'SSTL135',
    'SSTL15',
]


def gen_iosettings():
    """
    A generator function which yields all possible IO settings combintions.
    """

    IOSTANDARDS = (
        'LVCMOS12',
        'LVCMOS15',
        'LVCMOS18',
        'LVCMOS25',
        'LVCMOS33',
        'LVTTL',
        'SSTL135',
        'SSTL15',

        # Those are available but not currently fuzzed.
        #        'SSTL135_R',
        #        'SSTL15_R',
        #        'SSTL18_I',
        #        'SSTL18_II',
        #        'HSTL_I',
        #        'HSTL_I_18',
        #        'HSTL_II',
        #        'HSTL_II_18',
        #        'HSUL_12',
        #        'MOBILE_DDR',
    )

    DRIVES = defaultdict(lambda: [None])
    DRIVES.update(
        {
            "LVTTL": [4, 8, 12, 16, 24],
            "LVCMOS12": [4, 8, 12],
            "LVCMOS15": [4, 8, 12, 16],
            "LVCMOS18": [4, 8, 12, 16, 24],
            "LVCMOS25": [4, 8, 12, 16],
            "LVCMOS33": [4, 8, 12, 16],
        })

    SLEWS = ("SLOW", "FAST")

    for iostandard in IOSTANDARDS:

        # Single ended
        for drive in DRIVES[iostandard]:
            for slew in SLEWS:
                yield {"iostandard": iostandard, "drive": drive, "slew": slew}

        # Differential
        if iostandard in DIFF_MAP:
            for drive in DRIVES[iostandard]:
                for slew in SLEWS:
                    yield {
                        "iostandard": DIFF_MAP[iostandard],
                        "drive": drive,
                        "slew": slew
                    }


# =============================================================================


def run():
    """
    Main.
    """

    # Load IOB data
    iob_sites = load_iob_sites("iobs-{}.csv".format(os.getenv("PART")))

    # Generate IOB site to package pin map and *M site to *S site map.
    site_to_pkg_pin = {}
    master_to_slave = {}

    for region, sites in iob_sites.items():
        tiles = defaultdict(lambda: {})

        for site in sites:
            site_to_pkg_pin[site["site_name"]] = site["pkg_pin"]
            if site["site_type"] == "IOB33M":
                tiles[site["tile"]]["M"] = site
            if site["site_type"] == "IOB33S":
                tiles[site["tile"]]["S"] = site

        for sites in tiles.values():
            master_to_slave[sites["M"]["site_name"]] = sites["S"]["site_name"]

    # Generate designs
    iosettings_gen = gen_iosettings()
    design_index = 0
    while True:

        print("Design #{}".format(design_index))

        num_inp = 0
        num_out = 0
        num_ino = 0

        # Generate clock regions
        region_data = []
        for region in sorted(list(iob_sites.keys())):

            # Get IO bank. All sites from a clock region have the same one.
            bank = iob_sites[region][0]["bank"]

            # Get IO settings
            try:
                iosettings = next(iosettings_gen)
            except StopIteration:
                break

            # Get sites
            sites = [
                (
                    site["site_name"],
                    site["site_type"],
                )
                for site in iob_sites[region]
                if site["is_bonded"] and not int(site["is_vref"]) and "SING"
                not in site["tile"] and not "PUDC_B" in site["pin_func"]
            ]
            if not len(sites):
                continue

            # Differential / single ended
            if "DIFF" in iosettings["iostandard"]:

                # Select 5 random sites (IBUFDS, IBUFDS, OBUFDS, OBUFDS, IOBUFDS)
                site_names = [s[0] for s in sites if s[1] == "IOB33M"]
                used_sites = random.sample(site_names, 5)
                unused_sites = list(set(site_names) - set(used_sites))

                num_inp += 4
                num_out += 4
                num_ino += 2

            else:
                # Select 5 random sites (IBUF, IBUF, OBUF, OBUF, IOBUF)
                site_names = [s[0] for s in sites]
                used_sites = random.sample(site_names, 5)
                unused_sites = list(set(site_names) - set(used_sites))

                num_inp += 2
                num_out += 2
                num_ino += 1

            # Store data
            region_data.append(
                {
                    "region": region,
                    "bank": bank,
                    "iosettings": iosettings,
                    "unused_sites": unused_sites,
                    "input": used_sites[0:2],
                    "output": used_sites[2:4],
                    "inout": used_sites[4:5],
                })
            print("", region, iosettings)

        # No more
        if len(region_data) == 0:
            break

        print("----")

        # Generate the design
        verilog = """
module top (
    input  wire [{num_inp}:0] inp,
    inout  wire [{num_ino}:0] ino,
    output wire [{num_out}:0] out
);
""".format(num_inp=num_inp - 1, num_ino=num_ino - 1, num_out=num_out - 1)

        tcl = ""

        inp_idx = 0
        out_idx = 0
        ino_idx = 0

        for i, data in enumerate(region_data):

            is_diff = "DIFF" in data["iosettings"]["iostandard"]
            use_ino = data["iosettings"]["iostandard"] not in IOBUF_NOT_ALLOWED

            iostandard = data["iosettings"]["iostandard"]
            drive = data["iosettings"]["drive"]
            slew = data["iosettings"]["slew"]

            ibuf_param_str = ".IOSTANDARD(\"{}\")".format(iostandard)
            obuf_param_str = str(ibuf_param_str)

            if drive is not None:
                obuf_param_str += ", .DRIVE({})".format(drive)
            if slew is not None:
                obuf_param_str += ", .SLEW(\"{}\")".format(slew)

            bank = data["bank"]
            vref = "0.75"  # FIXME: Maybe loop over VREFs too ?

            keys = {
                "region": data["region"],
                "ibuf_0_loc": data["input"][0],
                "ibuf_1_loc": data["input"][1],
                "obuf_0_loc": data["output"][0],
                "obuf_1_loc": data["output"][1],
                "iobuf_loc": data["inout"][0],
                "inp_0_p": inp_idx,
                "inp_0_n": inp_idx + 2,
                "inp_1_p": inp_idx + 1,
                "inp_1_n": inp_idx + 3,
                "out_0_p": out_idx,
                "out_0_n": out_idx + 2,
                "out_1_p": out_idx + 1,
                "out_1_n": out_idx + 3,
                "ino_p": ino_idx,
                "ino_n": ino_idx + 1,
                "ibuf_param_str": ibuf_param_str,
                "obuf_param_str": obuf_param_str,
            }

            if is_diff:
                inp_idx += 4
                out_idx += 4
                ino_idx += 2
            else:
                inp_idx += 2
                out_idx += 2
                ino_idx += 1

            # Set VREF if necessary
            if iostandard in VREF_ALLOWED:
                tcl += "set_property INTERNAL_VREF {} [get_iobanks {}]\n".format(
                    vref, bank)

            # Single ended
            if not is_diff:

                tcl += "set_property PACKAGE_PIN {} [get_ports inp[{}]]\n".format(
                    site_to_pkg_pin[keys["ibuf_0_loc"]], keys["inp_0_p"])
                tcl += "set_property PACKAGE_PIN {} [get_ports inp[{}]]\n".format(
                    site_to_pkg_pin[keys["ibuf_1_loc"]], keys["inp_1_p"])
                tcl += "set_property PACKAGE_PIN {} [get_ports out[{}]]\n".format(
                    site_to_pkg_pin[keys["obuf_0_loc"]], keys["inp_0_p"])
                tcl += "set_property PACKAGE_PIN {} [get_ports out[{}]]\n".format(
                    site_to_pkg_pin[keys["obuf_1_loc"]], keys["inp_1_p"])
                tcl += "set_property PACKAGE_PIN {} [get_ports ino[{}]]\n".format(
                    site_to_pkg_pin[keys["iobuf_loc"]], keys["ino_p"])

                verilog += """

    // {region}
    wire inp_0_{region};
    wire inp_1_{region};
    wire out_0_{region};
    wire out_1_{region};

    wire ino_i_{region};
    wire ino_o_{region};
    wire ino_t_{region};
""".format(**keys)

                verilog += """
    (* KEEP, DONT_TOUCH *)
    IBUF # ({ibuf_param_str}) ibuf_0_{region} (
    .I(inp[{inp_0_p}]),
    .O(inp_0_{region})
    );

    (* KEEP, DONT_TOUCH *)
    IBUF # ({ibuf_param_str}) ibuf_1_{region} (
    .I(inp[{inp_1_p}]),
    .O(inp_1_{region})
    );

    (* KEEP, DONT_TOUCH *)
    OBUF # ({obuf_param_str}) obuf_0_{region} (
    .I(out_0_{region}),
    .O(out[{out_0_p}])
    );

    (* KEEP, DONT_TOUCH *)
    OBUF # ({obuf_param_str}) obuf_1_{region} (
    .I(out_1_{region}),
    .O(out[{out_1_p}])
    );
""".format(**keys)

                if use_ino:
                    verilog += """

    (* KEEP, DONT_TOUCH *)
    IOBUF # ({obuf_param_str}) iobuf_{region} (
    .I(ino_i_{region}),
    .O(ino_o_{region}),
    .T(ino_t_{region}),
    .IO(ino[{ino_p}])
    );

    assign out_0_{region} = inp_0_{region};
    assign out_1_{region} = ino_o_{region};
    assign ino_i_{region} = inp_0_{region};
    assign ino_t_{region} = inp_1_{region};
""".format(**keys)

                else:
                    verilog += """

    assign out_0_{region} = inp_0_{region};
    assign out_1_{region} = inp_1_{region};
    assign ino[{ino}] = 1'b0;
""".format(**keys)

            # Differential
            else:

                tcl += "set_property PACKAGE_PIN {} [get_ports inp[{}]]\n".format(
                    site_to_pkg_pin[keys["ibuf_0_loc"]], keys["inp_0_p"])
                tcl += "set_property PACKAGE_PIN {} [get_ports inp[{}]]\n".format(
                    site_to_pkg_pin[master_to_slave[keys["ibuf_0_loc"]]],
                    keys["inp_0_n"])
                tcl += "set_property PACKAGE_PIN {} [get_ports inp[{}]]\n".format(
                    site_to_pkg_pin[keys["ibuf_1_loc"]], keys["inp_1_p"])
                tcl += "set_property PACKAGE_PIN {} [get_ports inp[{}]]\n".format(
                    site_to_pkg_pin[master_to_slave[keys["ibuf_1_loc"]]],
                    keys["inp_1_n"])
                tcl += "set_property PACKAGE_PIN {} [get_ports out[{}]]\n".format(
                    site_to_pkg_pin[keys["obuf_0_loc"]], keys["inp_0_p"])
                tcl += "set_property PACKAGE_PIN {} [get_ports out[{}]]\n".format(
                    site_to_pkg_pin[master_to_slave[keys["obuf_0_loc"]]],
                    keys["inp_0_n"])
                tcl += "set_property PACKAGE_PIN {} [get_ports out[{}]]\n".format(
                    site_to_pkg_pin[keys["obuf_1_loc"]], keys["inp_1_p"])
                tcl += "set_property PACKAGE_PIN {} [get_ports out[{}]]\n".format(
                    site_to_pkg_pin[master_to_slave[keys["obuf_1_loc"]]],
                    keys["inp_1_n"])
                tcl += "set_property PACKAGE_PIN {} [get_ports ino[{}]]\n".format(
                    site_to_pkg_pin[keys["iobuf_loc"]], keys["ino_p"])
                tcl += "set_property PACKAGE_PIN {} [get_ports ino[{}]]\n".format(
                    site_to_pkg_pin[master_to_slave[keys["iobuf_loc"]]],
                    keys["ino_n"])

                verilog += """

    // {region}
    wire inp_0_{region};
    wire inp_1_{region};
    wire out_0_{region};
    wire out_1_{region};

    wire ino_i_{region};
    wire ino_o_{region};
    wire ino_t_{region};
""".format(**keys)

                verilog += """
    (* KEEP, DONT_TOUCH *)
    IBUFDS # ({ibuf_param_str}) ibufds_0_{region} (
    .I(inp[{inp_0_p}]),
    .IB(inp[{inp_0_n}]),
    .O(inp_0_{region})
    );

    (* KEEP, DONT_TOUCH *)
    IBUFDS # ({ibuf_param_str}) ibufds_1_{region} (
    .I(inp[{inp_1_p}]),
    .IB(inp[{inp_1_n}]),
    .O(inp_1_{region})
    );

    (* KEEP, DONT_TOUCH *)
    OBUFDS # ({obuf_param_str}) obufds_0_{region} (
    .I(out_0_{region}),
    .O(out[{out_0_p}]),
    .OB(out[{out_0_n}])
    );

    (* KEEP, DONT_TOUCH *)
    OBUFDS # ({obuf_param_str}) obufds_1_{region} (
    .I(out_1_{region}),
    .O(out[{out_1_p}]),
    .OB(out[{out_1_n}])
    );
""".format(**keys)

                if use_ino:
                    verilog += """

    (* KEEP, DONT_TOUCH *)
    IOBUFDS # ({obuf_param_str}) iobufds_{region} (
    .I(ino_i_{region}),
    .O(ino_o_{region}),
    .T(ino_t_{region}),
    .IO(ino[{ino_p}]),
    .IOB(ino[{ino_n}])
    );

    assign out_0_{region} = inp_0_{region};
    assign out_1_{region} = ino_o_{region};
    assign ino_i_{region} = inp_0_{region};
    assign ino_t_{region} = inp_1_{region};
""".format(**keys)

                else:
                    verilog += """

    assign out_0_{region} = inp_0_{region};
    assign out_1_{region} = inp_1_{region};
    assign ino[{ino_p}] = 1'b0;
    assign ino[{ino_n}] = 1'b0;
""".format(**keys)

        verilog += "endmodule"

        # Write verilog
        fname = "design_{:03d}.v".format(design_index)
        with open(fname, "w") as fp:
            fp.write(verilog)

        # Write TCL
        fname = "design_{:03d}.tcl".format(design_index)
        with open(fname, "w") as fp:
            fp.write(tcl)

        # Write JSON
        fname = "design_{:03d}.json".format(design_index)

        # Convert Vivado site names to fasm-style TILE.SITE names.
        for data in region_data:
            type_to_loc = {
                "IOB33": "IOB_Y0",
                "IOB33M": "IOB_Y0",
                "IOB33S": "IOB_Y1"
            }
            site_to_loc = {
                s["site_name"]: "{}.{}".format(
                    s["tile"], type_to_loc[s["site_type"]])
                for s in iob_sites[data["region"]]
            }

            for l in ["input", "output", "inout", "unused_sites"]:
                data[l] = [site_to_loc[s] for s in data[l]]

        # Write design settings to JSON
        with open(fname, "w") as fp:
            json.dump(region_data, fp, sort_keys=True, indent=1)

        design_index += 1


if __name__ == "__main__":
    run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import sys, os, re
from prjxray.util import OpenSafeFile, parse_db_lines, write_db_lines


def index_masks(fn_in, groups_in):
    """Return a dictionary with the bits active in each group for the specified list of groups"""
    # Only analyze the given groups
    groups = {}
    for group in groups_in:
        groups[group] = set()

    # Index bits
    for line, (tag, bits, mode) in parse_db_lines(fn_in):
        assert not mode, "Unresolved tag: %s" % (line, )
        prefix = tag[0:tag.rfind(".")]
        group = groups.get(prefix, None)
        # Drop groups we aren't interested in
        if group is None:
            continue
        for bit in bits:
            bit = bit.replace("!", "")
            group.add(bit)

    # Verify we were able to find all groups
    for groupk, groupv in groups.items():
        assert len(groupv), "Bad group %s" % groupk

    return groups


def apply_masks(fn_in, groups):
    """Add 0 entries ("!") to .db entries based on groups definition"""
    new_db = {}
    for line, (tag, bits, mode) in parse_db_lines(fn_in):
        assert not mode, "Unresolved tag: %s" % (line, )
        prefix = tag[0:tag.rfind(".")]
        group = groups.get(prefix, None)
        if group:
            bits = set(bits)
            for bit in group:
                if bit not in bits:
                    bits.add("!" + bit)
            bits = frozenset(bits)
        new_db[tag] = bits
    return new_db


def load_groups(fn):
    ret = []
    with OpenSafeFile(fn, "r") as f:
        for l in f:
            ret.append(l.strip())
    return ret


def run(fn_in, fn_out, groups_fn, verbose=False):
    groups_in = load_groups(groups_fn)
    groups = index_masks(fn_in, groups_in)
    new_db = apply_masks(fn_in, groups)
    write_db_lines(fn_out, new_db)


def main():
    import argparse

    parser = argparse.ArgumentParser(description='Create multi-bit entries')
    parser.add_argument('--verbose', action='store_true', help='')
    parser.add_argument(
        '--groups-fn',
        default="groups.grp",
        help='File containing one group per line to parse')
    parser.add_argument('fn_in', help='')
    parser.add_argument('fn_out', help='')
    args = parser.parse_args()

    run(args.fn_in, args.fn_out, args.groups_fn, verbose=args.verbose)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

from prjxray.segmaker import Segmaker

segmk = Segmaker("design.bits")

print("Loading tags from design.txt.")
with open("design.txt", "r") as f:
    for line in f:
        line, active = line.split()
        tile, pip = line.split("/")
        _, pip = pip.split(".")

        print(tile, pip, active)
        segmk.addtag(tile, pip, int(active))

segmk.compile()
segmk.write()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
"""
This script loads the PS7 pin dump from Vivado and groups pins into ports. Also
assigns each port a class that indicates its function. The classes are:

- "normal": A port that connects to the PL (FPGA)
- "test":   A port used for testing, not accessible from the PL.
- "debug":  A debug port, not accessible.
- "mio":    The "mio" ports go directly to die pads, not relevant for routing.

Ports are then written to a JSON file.
"""
import argparse
import csv
import json
import re

from collections import defaultdict

# =============================================================================


def main():

    BUS_REGEX = re.compile("(.*[A-Z_])([0-9]+)$")

    # Parse arguments
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument("csv", type=str, help="PS7 pin dump file")
    parser.add_argument(
        "json",
        type=str,
        help="Output JSON file with PS7 pins grouped into ports")

    args = parser.parse_args()

    # Load pin dump
    with open(args.csv, "r") as fp:
        pin_dump = list(csv.DictReader(fp))

    # Group pins into ports
    ports = defaultdict(lambda :{
        "direction": None,
        "min": None,
        "max": None,
        "width": 0
        })

    for pin in list(pin_dump):

        # Get port name and signal index
        match = BUS_REGEX.match(pin["name"])
        if match:
            name = match.group(1)
            idx = int(match.group(2))
        else:
            name = pin["name"]
            idx = 0

        # Get direction
        is_input = int(pin["is_input"])
        is_output = int(pin["is_output"])
        is_bidir = int(pin["is_bidir"])

        if is_input and not is_output and not is_bidir:
            direction = "input"
        elif not is_input and is_output and not is_bidir:
            direction = "output"
        elif not is_input and not is_output and is_bidir:
            direction = "inout"
        else:
            assert False, pin

        # Add to port
        port = ports[name]

        if port["direction"] is None:
            port["direction"] = direction
        else:
            assert port["direction"] == direction

        if port["min"] is None:
            port["min"] = idx
        else:
            port["min"] = min(port["min"], idx)

        if port["max"] is None:
            port["max"] = idx
        else:
            port["max"] = max(port["max"], idx)

        port["width"] = port["max"] - port["min"] + 1

    # Sort ports by their purpose
    for name, port in ports.items():

        # A test pin (unconnected)
        if name.startswith("TEST"):
            cls = "test"

        # A debug pin (unconnected)
        elif name.startswith("DEBUG"):
            cls = "debug"

        # A MIO/DDR pin.
        elif name.startswith("MIO") or name.startswith("DDR") and \
            name != "DDRARB":
            cls = "mio"

        # PS7 clock/reset
        elif name in ["PSCLK", "PSPORB", "PSSRSTB"]:
            cls = "mio"

        # "Normal" pin
        else:
            cls = "normal"

        port["class"] = cls

    # Write pin ports to a JSON file
    with open(args.json, "w") as fp:
        json.dump(ports, fp, indent=1, sort_keys=True)


# =============================================================================

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import os
import random
random.seed(int(os.getenv("SEED"), 16))
from prjxray import util
from prjxray.db import Database


def gen_sites():
    '''
    IOB18S: main IOB of a diff pair
    IOB18M: secondary IOB of a diff pair
    IOB18: not a diff pair. Relatively rare (at least in ROI...2 of them?)
    Focus on IOB18S to start
    '''
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()
    for tile_name in sorted(grid.tiles()):
        loc = grid.loc_of_tilename(tile_name)
        gridinfo = grid.gridinfo_at_loc(loc)

        for site_name, site_type in gridinfo.sites.items():
            if site_type == 'IOB18S':
                yield tile_name, site_name


def write_params(params):
    pinstr = 'tile,val,site,pin\n'
    for tile, (site, val, pin) in sorted(params.items()):
        pinstr += '%s,%s,%s,%s\n' % (tile, val, site, pin)
    open('params.csv', 'w').write(pinstr)


def run():
    sites = list(gen_sites())
    print(
        '''
`define N_DI {}

module top(input wire [`N_DI-1:0] di);
    wire [`N_DI-1:0] di_buf;
    '''.format(len(sites)))

    params = {}
    print('''
        (* KEEP, DONT_TOUCH *)
        LUT6 dummy_lut();''')

    for idx, ((tile_name, site_name), isone) in enumerate(zip(
            sites, util.gen_fuzz_states(len(sites)))):
        params[tile_name] = (site_name, isone, "di[%u]" % idx)
        print(
            '''
    (* KEEP, DONT_TOUCH *)
    IBUF #(
    ) ibuf_{site_name} (
        .I(di[{idx}]),
        .O(di_buf[{idx}])
        );'''.format(site_name=site_name, idx=idx))

        if isone:
            print(
                '''
    (* KEEP, DONT_TOUCH *)
    PULLUP #(
    ) pullup_{site_name} (
        .O(di[{idx}])
        );'''.format(site_name=site_name, idx=idx))

    print("endmodule")
    write_params(params)


if __name__ == '__main__':
    run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import os.path
import pathlib
import simplejson as json
from prjxray import grid
from prjxray import tile
from prjxray import tile_segbits
from prjxray import site_type
from prjxray import connections
from prjxray.node_model import NodeModel
from prjxray.util import get_fabric_for_part


def get_available_databases(prjxray_root):
    """ Return set of available directory to databases given the root directory
      of prjxray-db
  """
    db_types = set()
    for d in os.listdir(prjxray_root):
        if d.startswith("."):
            continue

        dpath = os.path.join(prjxray_root, d)

        if os.path.exists(os.path.join(dpath, "settings.sh")):
            db_types.add(dpath)

    return db_types


class Database(object):
    def __init__(self, db_root, part):
        """ Create project x-ray Database at given db_root.

    db_root: Path to directory containing settings.sh, *.db, tilegrid.json and
             tileconn.json

    """
        self.db_root = db_root
        self.part = part
        self.fabric = get_fabric_for_part(db_root, part)

        # tilegrid.json JSON object
        self.tilegrid = None
        self.tileconn = None
        self.tile_types_json = None
        self.node_wires = None

        self.tile_types = {}
        self.tile_segbits = {}
        self.site_types = {}

        self.required_features = {}

        for f in os.listdir(self.db_root):
            if f.endswith('.json') and f.startswith('tile_type_'):
                tile_type = f[len('tile_type_'):-len('.json')].lower()

                segbits = os.path.join(
                    self.db_root, 'segbits_{}.db'.format(tile_type))
                if not os.path.isfile(segbits):
                    segbits = None

                block_ram_segbits = os.path.join(
                    self.db_root, 'segbits_{}.block_ram.db'.format(tile_type))
                if not os.path.isfile(block_ram_segbits):
                    block_ram_segbits = None

                ppips = os.path.join(
                    self.db_root, 'ppips_{}.db'.format(tile_type))
                if not os.path.isfile(ppips):
                    ppips = None

                mask = os.path.join(
                    self.db_root, 'mask_{}.db'.format(tile_type))
                if not os.path.isfile(mask):
                    mask = None

                tile_type_file = os.path.join(
                    self.db_root, 'tile_type_{}.json'.format(
                        tile_type.upper()))
                if not os.path.isfile(tile_type_file):
                    tile_type_file = None

                self.tile_types[tile_type.upper()] = tile.TileDbs(
                    segbits=segbits,
                    block_ram_segbits=block_ram_segbits,
                    ppips=ppips,
                    mask=mask,
                    tile_type=tile_type_file,
                )

            if f.endswith('.json') and f.startswith('site_type_'):
                site_type_name = f[len('site_type_'):-len('.json')]

                self.site_types[site_type_name] = os.path.join(self.db_root, f)

        required_features_path = os.path.join(
            self.db_root, self.part, "required_features.fasm")
        if os.path.isfile(required_features_path):
            with open(required_features_path, "r") as fp:
                features = []
                for line in fp:
                    line = line.strip()
                    if len(line) > 0:
                        features.append(line)

                self.required_features[self.part] = set(features)

        self.tile_types_obj = {}

    def get_tile_types(self):
        """ Return list of tile types """
        return self.tile_types.keys()

    def get_tile_type(self, tile_type):
        """ Return Tile object for given tilename. """
        if tile_type not in self.tile_types_obj:
            self.tile_types_obj[tile_type] = tile.Tile(
                tile_type, self.tile_types[tile_type])

        return self.tile_types_obj[tile_type]

    def _read_tilegrid(self):
        """ Read tilegrid database if not already read. """
        if not self.tilegrid:
            with open(os.path.join(self.db_root, self.fabric,
                                   'tilegrid.json')) as f:
                self.tilegrid = json.load(f)

    def _read_tileconn(self):
        """ Read tileconn database if not already read. """
        if not self.tileconn:
            with open(os.path.join(self.db_root, self.fabric,
                                   'tileconn.json')) as f:
                self.tileconn = json.load(f)

    def _read_node_wires(self):
        """ Read node wires if not already read. """
        if self.node_wires is None:
            with open(os.path.join(self.db_root, self.fabric,
                                   'node_wires.json')) as f:
                self.node_wires = json.load(f)

    def grid(self):
        """ Return Grid object for database. """
        self._read_tilegrid()
        return grid.Grid(self, self.tilegrid)

    def _read_tile_types(self):
        if self.tile_types_json is None:
            self.tile_types_json = {}
            for tile_type, db in self.tile_types.items():
                with open(db.tile_type) as f:
                    self.tile_types_json[tile_type] = json.load(f)

    def _get_tile_wires(self):
        self._read_tile_types()
        tile_wires = dict(
            (tile_type, db['wires'])
            for tile_type, db in self.tile_types_json.items())

        return tile_wires

    def connections(self):
        self._read_tilegrid()
        self._read_tileconn()

        return connections.Connections(
            self.tilegrid, self.tileconn, self._get_tile_wires())

    def node_model(self, progressbar=lambda x: x):
        """ Get node module for specified part.

        progressbar - Should be a function that takes an iteraable, and
                      yields all elements from that iterable.
                      This can be used to generate a progressbar, for example
                      the module progressbar satifies this interface.

                        Example:

                        import progressbar

                        db = Database(...)
                        node_model = db.node_model(progressbar.progressbar)

        """
        self._read_node_wires()

        return NodeModel(
            grid=self.grid(),
            connections=self.connections(),
            tile_wires=self._get_tile_wires(),
            node_wires=self.node_wires,
            progressbar=progressbar)

    def get_site_types(self):
        return self.site_types.keys()

    def get_site_type(self, site_type_name):
        with open(self.site_types[site_type_name]) as f:
            site_type_data = json.load(f)

        return site_type.SiteType(site_type_data)

    def get_tile_segbits(self, tile_type):
        if tile_type not in self.tile_segbits:
            self.tile_segbits[tile_type] = tile_segbits.TileSegbits(
                self.tile_types[tile_type.upper()])

        return self.tile_segbits[tile_type]

    def get_required_fasm_features(self, part=None):
        """
        Assembles a set of required fasm features for given part. Returns a list
        of fasm features.
        """

        # No required features in the db, return empty list
        if self.required_features is None:
            return set()

        # Return list of part specific features
        return self.required_features.get(part, set())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
'''
Prints minor address group sizes
If any part of an IP block is written, Vivado likes to write the entire block
This can be useful to guess the number of frames an IP block uses

This was developed with PERFRAMECRC, but could probably be rewritten to work without it
'''

import sys
import os
import time
import json
import subprocess
from io import StringIO

BITTOOL = os.getenv('XRAY_BITTOOL')


def bit2packets(fn):
    # capture_output wasn't added until 3.7, I have 3.5
    #return subprocess.run([BITTOOL, "list_config_packets", fn], capture_output=True, check=True).stdout
    p = subprocess.run(
        [BITTOOL, "list_config_packets", fn],
        check=True,
        stdout=subprocess.PIPE)
    return p.stdout.decode("ascii")


def nominor(addr):
    return addr & ~0x7f


def gen_frame_writes(f):
    '''
    look for line triples like this

    [Write Type=1 Address= 1 Length=         1 Reg="Frame Address"]
    Data in hex:
          1d

    '''
    while True:
        l = f.readline()
        if not l:
            break
        l = l.strip()
        # TODO: assert that writes are always per frame CRC
        if l != '[Write Type=1 Address= 1 Length=         1 Reg="Frame Address"]':
            continue
        assert f.readline().strip() == 'Data in hex:'
        lhex = f.readline().strip()
        yield int(lhex, 16)


def gen_major_writes(fnin):
    '''
    The same address can appear twice
    Ex: 0x00800000
    Moved from counter to set
    '''
    last_addrs = None
    last_major = None

    for this_minor in gen_frame_writes(StringIO(bit2packets(fnin))):
        this_major = nominor(this_minor)

        if this_major == last_major:
            # should be no gaps
            last_addrs.add(this_minor)

            # addresses may skip back, but they don't seem to skip forward
            # AssertionError: (0, 2, {0, 1})
            # assert (this_major - this_minor) + 1 == len(last_addrs), (this_major, len(last_addrs), last_addrs)
            assert (this_major - this_minor) + 1 <= len(last_addrs), (
                this_major, len(last_addrs), last_addrs)
        else:
            if last_major is not None:
                assert len(last_addrs) <= 0x80
                yield last_major, len(last_addrs)

            # all writes should start at base?
            assert this_major == this_minor
            last_major = this_major
            last_addrs = set([this_minor])
    yield last_major, len(last_addrs)


def run(fnin, verbose=False):
    for addr, nframes in gen_major_writes(fnin):
        print('0x%08X: 0x%02X' % (addr, nframes))


def main():
    import argparse

    parser = argparse.ArgumentParser(
        description='Print PERFRAMECRC bitstream minor address gropuing')
    parser.add_argument('fnin', default=None, help='input bitstream')
    args = parser.parse_args()

    run(args.fnin, verbose=False)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import json

from prjxray.segmaker import Segmaker


def bitfilter(frame, bit):
    # Filter out interconnect bits.
    if frame in [0, 1, 17, 23, 25]:
        return False

    # ZINV_RSTRAMARSTRAM seems hard to de-correlate from FIFO_Y0_IN_USE.
    if (frame, bit) == (27, 116):
        return False

    return True


def main():
    segmk = Segmaker("design.bits")

    print("Loading tags")
    with open('params.json') as f:
        params = json.load(f)

    for tile_param in params:
        for param, tag in (
            ('Y0_IN_USE', 'RAMB18_Y0.IN_USE'),
            ('Y1_IN_USE', 'RAMB18_Y1.IN_USE'),
            ('FIFO_Y0_IN_USE', 'RAMB18_Y0.FIFO_MODE'),
            ('FIFO_Y1_IN_USE', 'RAMB18_Y1.FIFO_MODE'),
        ):
            segmk.add_tile_tag(tile_param['tile'], tag, tile_param[param])

    segmk.compile(bitfilter=bitfilter)
    segmk.write()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import sys
import os
sys.path.append(os.getenv("XRAY_UTILS_DIR"))
import simpleroute

print()
print('ready')


def load_design(f):
    '''
    name node pin wire
    clk CLK_HROW_TOP_R_X60Y130/CLK_HROW_CK_BUFHCLK_L0 W5 HCLK_VBRK_X34Y130/HCLK_VBRK_CK_BUFHCLK0
    din[0] INT_R_X9Y100/NE2BEG3 V17 VBRK_X29Y106/VBRK_NE2A3
    '''
    ret = {}
    f.readline()
    for l in f:
        l = l.strip()
        name, node, pin, wire = l.split(' ')
        ret[name] = wire
    return ret


def route2fasm(route, out_f):
    pips = simpleroute.route(route)
    for pip in pips:
        # INT_L_X10Y122.NL1BEG2.NE2END3
        # to
        # INT_L_X10Y122.NL1BEG2 NE2END3
        doti = pip.rfind('.')
        pip = pip[0:doti] + ' ' + pip[doti + 1:]
        out_f.write(pip + '\n')


def run(design_f, swn, ledn, out_f):
    name2wire = load_design(design_f)
    led_name = 'dout[%d]' % ledn
    sw_name = 'din[%d]' % swn
    led_wire = name2wire[led_name]
    sw_wire = name2wire[sw_name]
    print(
        'Routing %s (%s) => %s (%s)' % (sw_wire, sw_name, led_wire, led_name))

    route2fasm((sw_wire, led_wire), out_f)
    # XXX: terminate LEDs so they are off?


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(
        description='Switch to LED interconnect demo: FASM generator')
    parser.add_argument('design_txt', help='ROI metadata file')
    parser.add_argument('sw', type=int, help='Switch to use')
    parser.add_argument('led', type=int, help='LED to use')
    # For now can't use stdout since simpleroute is spewing out prints
    parser.add_argument('out_fasm', help='Output .fasm file')

    args = parser.parse_args()
    run(
        open(args.design_txt, 'r'), args.sw, args.led, open(
            args.out_fasm, 'w'))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import os, random
random.seed(0)
from prjxray import util
from prjxray import verilog

from prims import *

# INCREMENT is the amount of additional CLBN to be instantiated in the design.
# This makes the fuzzer compilation more robust against failures.
INCREMENT = os.getenv('CLBN', 0)
CLBN = 600 + int(INCREMENT)
print('//Requested CLBs: %s' % str(CLBN))

f = open("top.txt", "w")
f.write("i,prim,loc,bel,init\n")


def gen_slices():
    for _tile_name, site_name, _site_type in util.get_roi().gen_sites([
            'SLICEL',
            'SLICEM',
    ]):
        yield site_name


DIN_N = CLBN * 4
DOUT_N = CLBN * 1

verilog.top_harness(DIN_N, DOUT_N)

slices = gen_slices()
print(
    'module roi(input clk, input [%d:0] din, output [%d:0] dout);' %
    (DIN_N - 1, DOUT_N - 1))
for i in range(CLBN):
    ffprim = random.choice(ones(ffprims))
    # clb_FD clb_FD (.clk(clk), .din(din[  0 +: 4]), .dout(dout[  0]));
    # clb_FD_1 clb_FD_1 (.clk(clk), .din(din[  4 +: 4]), .dout(dout[  1]));
    loc = next(slices)
    # Latch can't go in 5s
    if isff(ffprim):
        bel = random.choice(ff_bels)
    else:
        bel = random.choice(ff_bels_ffl)
    init = random.choice((0, 1))
    #bel = "AFF"
    print('    clb_%s' % ffprim)
    print(
        '            #(.LOC("%s"), .BEL("%s"), .INIT(%d))' % (loc, bel, init))
    print(
        '            clb_%d (.clk(clk), .din(din[  %d +: 4]), .dout(dout[  %d]));'
        % (i, 4 * i, 1 * i))
    f.write("%d,%s,%s,%s,%d\n" % (i, ffprim, loc, bel, init))
print(
    '''endmodule

// ---------------------------------------------------------------------

''')

print(
    '''
module clb_FD (input clk, input [3:0] din, output dout);
    parameter LOC="SLICE_X16Y100";
    parameter BEL="AFF";
    parameter INIT=1'b0;
	(* LOC=LOC, BEL=BEL, KEEP, DONT_TOUCH, INIT=INIT *)
	FD ff (
		.C(clk),
		.Q(dout),
		.D(din[0])
	);
	endmodule

module clb_FD_1 (input clk, input [3:0] din, output dout);
    parameter LOC="SLICE_X16Y101";
    parameter BEL="AFF";
    parameter INIT=1'b0;
	(* LOC=LOC, BEL=BEL, KEEP, DONT_TOUCH, INIT=INIT *)
	FD_1 ff (
		.C(clk),
		.Q(dout),
		.D(din[0])
	);
	endmodule

module clb_FDC (input clk, input [3:0] din, output dout);
    parameter LOC="SLICE_X16Y102";
    parameter BEL="AFF";
    parameter INIT=1'b0;
	(* LOC=LOC, BEL=BEL, KEEP, DONT_TOUCH, INIT=INIT *)
	FDC ff (
		.C(clk),
		.Q(dout),
		.CLR(din[0]),
		.D(din[1])
	);
	endmodule

module clb_FDC_1 (input clk, input [3:0] din, output dout);
    parameter LOC="SLICE_X16Y103";
    parameter BEL="AFF";
    parameter INIT=1'b0;
	(* LOC=LOC, BEL=BEL, KEEP, DONT_TOUCH, INIT=INIT *)
	FDC_1 ff (
		.C(clk),
		.Q(dout),
		.CLR(din[0]),
		.D(din[1])
	);
	endmodule

module clb_FDCE (input clk, input [3:0] din, output dout);
    parameter LOC="SLICE_X16Y104";
    parameter BEL="AFF";
    parameter INIT=1'b0;
	(* LOC=LOC, BEL=BEL, KEEP, DONT_TOUCH, INIT=INIT *)
	FDCE ff (
		.C(clk),
		.Q(dout),
		.CE(din[0]),
		.CLR(din[1]),
		.D(din[2])
	);
	endmodule

module clb_FDCE_1 (input clk, input [3:0] din, output dout);
    parameter LOC="SLICE_X16Y105";
    parameter BEL="AFF";
    parameter INIT=1'b0;
	(* LOC=LOC, BEL=BEL, KEEP, DONT_TOUCH, INIT=INIT *)
	FDCE_1 ff (
		.C(clk),
		.Q(dout),
		.CE(din[0]),
		.CLR(din[1]),
		.D(din[2])
	);
	endmodule

module clb_FDE (input clk, input [3:0] din, output dout);
    parameter LOC="SLICE_X16Y106";
    parameter BEL="AFF";
    parameter INIT=1'b0;
	(* LOC=LOC, BEL=BEL, KEEP, DONT_TOUCH, INIT=INIT *)
	FDE ff (
		.C(clk),
		.Q(dout),
		.D(din[0]),
		.CE(din[1])
	);
	endmodule

module clb_FDE_1 (input clk, input [3:0] din, output dout);
    parameter LOC="SLICE_X16Y107";
    parameter BEL="AFF";
    parameter INIT=1'b0;
	(* LOC=LOC, BEL=BEL, KEEP, DONT_TOUCH, INIT=INIT *)
	FDE_1 ff (
		.C(clk),
		.Q(dout),
		.D(din[0]),
		.CE(din[1])
	);
	endmodule

module clb_FDP (input clk, input [3:0] din, output dout);
    parameter LOC="SLICE_X16Y108";
    parameter BEL="AFF";
    parameter INIT=1'b0;
	(* LOC=LOC, BEL=BEL, KEEP, DONT_TOUCH, INIT=INIT *)
	FDP ff (
		.C(clk),
		.Q(dout),
		.D(din[0]),
		.PRE(din[1])
	);
	endmodule

module clb_FDP_1 (input clk, input [3:0] din, output dout);
    parameter LOC="SLICE_X16Y109";
    parameter BEL="AFF";
    parameter INIT=1'b0;
	(* LOC=LOC, BEL=BEL, KEEP, DONT_TOUCH, INIT=INIT *)
	FDP_1 ff (
		.C(clk),
		.Q(dout),
		.D(din[0]),
		.PRE(din[1])
	);
	endmodule

module clb_FDPE (input clk, input [3:0] din, output dout);
    parameter LOC="SLICE_X16Y110";
    parameter BEL="AFF";
    parameter INIT=1'b0;
	(* LOC=LOC, BEL=BEL, KEEP, DONT_TOUCH, INIT=INIT *)
	FDPE ff (
		.C(clk),
		.Q(dout),
		.CE(din[0]),
		.PRE(din[1]),
		.D(din[2])
	);
	endmodule

module clb_FDPE_1 (input clk, input [3:0] din, output dout);
    parameter LOC="SLICE_X16Y111";
    parameter BEL="AFF";
    parameter INIT=1'b0;
	(* LOC=LOC, BEL=BEL, KEEP, DONT_TOUCH, INIT=INIT *)
	FDPE_1 ff (
		.C(clk),
		.Q(dout),
		.CE(din[0]),
		.PRE(din[1]),
		.D(din[2])
	);
	endmodule

module clb_FDR (input clk, input [3:0] din, output dout);
    parameter LOC="SLICE_X16Y112";
    parameter BEL="AFF";
    parameter INIT=1'b0;
	(* LOC=LOC, BEL=BEL, KEEP, DONT_TOUCH, INIT=INIT *)
	FDR ff (
		.C(clk),
		.Q(dout),
		.D(din[0]),
		.R(din[1])
	);
	endmodule

module clb_FDR_1 (input clk, input [3:0] din, output dout);
    parameter LOC="SLICE_X16Y113";
    parameter BEL="AFF";
    parameter INIT=1'b0;
	(* LOC=LOC, BEL=BEL, KEEP, DONT_TOUCH, INIT=INIT *)
	FDR_1 ff (
		.C(clk),
		.Q(dout),
		.D(din[0]),
		.R(din[1])
	);
	endmodule

module clb_FDRE (input clk, input [3:0] din, output dout);
    parameter LOC="SLICE_X16Y114";
    parameter BEL="AFF";
    parameter INIT=1'b0;
	(* LOC=LOC, BEL=BEL, KEEP, DONT_TOUCH, INIT=INIT *)
	FDRE ff (
		.C(clk),
		.Q(dout),
		.CE(din[0]),
		.R(din[1]),
		.D(din[2])
	);
	endmodule

module clb_FDRE_1 (input clk, input [3:0] din, output dout);
    parameter LOC="SLICE_X16Y115";
    parameter BEL="AFF";
    parameter INIT=1'b0;
	(* LOC=LOC, BEL=BEL, KEEP, DONT_TOUCH, INIT=INIT *)
	FDRE_1 ff (
		.C(clk),
		.Q(dout),
		.CE(din[0]),
		.R(din[1]),
		.D(din[2])
	);
	endmodule

module clb_FDS (input clk, input [3:0] din, output dout);
    parameter LOC="SLICE_X16Y116";
    parameter BEL="AFF";
    parameter INIT=1'b0;
	(* LOC=LOC, BEL=BEL, KEEP, DONT_TOUCH, INIT=INIT *)
	FDS ff (
		.C(clk),
		.Q(dout),
		.S(din[0]),
		.D(din[1])
	);
	endmodule

module clb_FDS_1 (input clk, input [3:0] din, output dout);
    parameter LOC="SLICE_X16Y117";
    parameter BEL="AFF";
    parameter INIT=1'b0;
	(* LOC=LOC, BEL=BEL, KEEP, DONT_TOUCH, INIT=INIT *)
	FDS_1 ff (
		.C(clk),
		.Q(dout),
		.S(din[0]),
		.D(din[1])
	);
	endmodule

module clb_FDSE (input clk, input [3:0] din, output dout);
    parameter LOC="SLICE_X16Y118";
    parameter BEL="AFF";
    parameter INIT=1'b0;
	(* LOC=LOC, BEL=BEL, KEEP, DONT_TOUCH, INIT=INIT *)
	FDSE ff (
		.C(clk),
		.Q(dout),
		.CE(din[0]),
		.S(din[1]),
		.D(din[2])
	);
	endmodule

module clb_FDSE_1 (input clk, input [3:0] din, output dout);
    parameter LOC="SLICE_X16Y119";
    parameter BEL="AFF";
    parameter INIT=1'b0;
	(* LOC=LOC, BEL=BEL, KEEP, DONT_TOUCH, INIT=INIT *)
	FDSE_1 ff (
		.C(clk),
		.Q(dout),
		.CE(din[0]),
		.S(din[1]),
		.D(din[2])
	);
	endmodule

//********************************************************************************

module clb_LDC (input clk, input [3:0] din, output dout);
    parameter LOC="SLICE_X16Y120";
    parameter BEL="AFF";
    parameter INIT=1'b0;
	(* LOC=LOC, BEL=BEL, KEEP, DONT_TOUCH, INIT=INIT *)
	LDC ff (
		.G(~clk),
		.Q(dout),
		.D(din[0]),
		.CLR(din[1])
	);
	endmodule
module clb_LDC_1 (input clk, input [3:0] din, output dout);
    parameter LOC="SLICE_X16Y121";
    parameter BEL="AFF";
    parameter INIT=1'b0;
	(* LOC=LOC, BEL=BEL, KEEP, DONT_TOUCH, INIT=INIT *)
	LDC_1 ff (
		.G(~clk),
		.Q(dout),
		.D(din[0]),
		.CLR(din[1])
	);
	endmodule

module clb_LDCE (input clk, input [3:0] din, output dout);
    parameter LOC="SLICE_X16Y122";
    parameter BEL="AFF";
    parameter INIT=1'b0;
	(* LOC=LOC, BEL=BEL, KEEP, DONT_TOUCH, INIT=INIT *)
	LDCE ff (
		.G(~clk),
		//NOTE: diagram shows two outputs. Error?
		.Q(dout),
		.D(din[0]),
		.GE(din[1]),
		.CLR(din[2])
	);
	endmodule
module clb_LDCE_1 (input clk, input [3:0] din, output dout);
    parameter LOC="SLICE_X16Y123";
    parameter BEL="AFF";
    parameter INIT=1'b0;
	(* LOC=LOC, BEL=BEL, KEEP, DONT_TOUCH, INIT=INIT *)
	LDCE_1 ff (
		.G(~clk),
		//NOTE: diagram shows two outputs. Error?
		.Q(dout),
		.D(din[0]),
		.GE(din[1]),
		.CLR(din[2])
	);
	endmodule

module clb_LDE (input clk, input [3:0] din, output dout);
    parameter LOC="SLICE_X16Y124";
    parameter BEL="AFF";
    parameter INIT=1'b0;
	(* LOC=LOC, BEL=BEL, KEEP, DONT_TOUCH, INIT=INIT *)
	LDE ff (
		.G(~clk),
		.Q(dout),
		.D(din[0]),
		.GE(din[1])
	);
	endmodule
module clb_LDE_1 (input clk, input [3:0] din, output dout);
    parameter LOC="SLICE_X16Y125";
    parameter BEL="AFF";
    parameter INIT=1'b0;
	(* LOC=LOC, BEL=BEL, KEEP, DONT_TOUCH, INIT=INIT *)
	LDE_1 ff (
		.G(~clk),
		.Q(dout),
		.D(din[0]),
		.GE(din[1])
	);
	endmodule

module clb_LDP (input clk, input [3:0] din, output dout);
    parameter LOC="SLICE_X16Y126";
    parameter BEL="AFF";
    parameter INIT=1'b0;
	(* LOC=LOC, BEL=BEL, KEEP, DONT_TOUCH, INIT=INIT *)
	LDP ff (
		.G(~clk),
		.Q(dout),
		.D(din[0]),
		.PRE(din[1])
	);
	endmodule
module clb_LDP_1 (input clk, input [3:0] din, output dout);
    parameter LOC="SLICE_X16Y127";
    parameter BEL="AFF";
    parameter INIT=1'b0;
	(* LOC=LOC, BEL=BEL, KEEP, DONT_TOUCH, INIT=INIT *)
	LDP_1 ff (
		.G(~clk),
		.Q(dout),
		.D(din[0]),
		.PRE(din[1])
	);
	endmodule

module clb_LDPE (input clk, input [3:0] din, output dout);
    parameter LOC="SLICE_X16Y128";
    parameter BEL="AFF";
    parameter INIT=1'b0;
	(* LOC=LOC, BEL=BEL, KEEP, DONT_TOUCH, INIT=INIT *)
	LDPE ff (
		.G(~clk),
		.Q(dout),
		.PRE(din[0]),
		.D(din[1]),
		.GE(din[2])
	);
	endmodule
module clb_LDPE_1 (input clk, input [3:0] din, output dout);
    parameter LOC="SLICE_X16Y129";
    parameter BEL="AFF";
    parameter INIT=1'b0;
	(* LOC=LOC, BEL=BEL, KEEP, DONT_TOUCH, INIT=INIT *)
	LDPE_1 ff (
		.G(~clk),
		.Q(dout),
		.PRE(din[0]),
		.D(din[1]),
		.GE(din[2])
	);
	endmodule

''')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

from prjxray import bitsmaker


def run(
        bits_fn,
        design_fn,
        fnout,
        oneval,
        dframe,
        auto_frame,
        dword,
        dbit,
        multi=False,
        verbose=False):
    # mimicing tag names, wasn't sure if it would break things otherwise
    metastr = "DWORD:%u" % dword
    if dbit is not None:
        metastr += ".DBIT:%u" % dbit
    if dframe is not None:
        metastr += ".DFRAME:%02x" % dframe
    if multi:
        metastr += ".MULTI"
    if auto_frame:
        metastr += ".AUTO_FRAME"

    tags = dict()
    f = open(design_fn, 'r')
    f.readline()
    for l in f:
        l = l.strip()
        # Additional values reserved / for debugging
        tile, val = l.split(',')[0:2]
        tags["%s.%s" % (tile, metastr)] = val == oneval

    bitsmaker.write(bits_fn, fnout, tags)


def main():
    import argparse

    parser = argparse.ArgumentParser(
        description=
        "Solve bits (like segmaker) on raw .bits file without segments")
    parser.add_argument("--bits-file", default="design.bits", help="")
    parser.add_argument("--verbose", action="store_true", help="")
    parser.add_argument("--design", default="design.csv", help="")
    parser.add_argument("--fnout", default="/dev/stdout", help="")
    parser.add_argument(
        "--oneval",
        required=True,
        help="Parameter value that correspodns to a set bit")
    #
    parser.add_argument(
        "--multi", action="store_true", help="Are multiple bits expected?")
    parser.add_argument(
        "--dframe",
        type=str,
        required=False,
        default="",
        help="Reference frame delta (base 16)")
    parser.add_argument(
        "--auto-frame",
        action='store_true',
        help="Auto align frame address to next lowest multiple of 0x80")
    parser.add_argument(
        "--dword",
        type=str,
        required=True,
        help="Reference word delta (base 10)")
    parser.add_argument(
        "--dbit",
        type=str,
        required=False,
        default="",
        help="Reference bit delta (base 10)")
    args = parser.parse_args()

    run(
        args.bits_file,
        args.design,
        args.fnout,
        args.oneval,
        None if args.dframe == "" else int(args.dframe, 16),
        args.auto_frame,
        int(args.dword, 10),
        None if args.dbit == "" else int(args.dbit, 10),
        multi=args.multi,
        verbose=args.verbose)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

from __future__ import print_function

import os

from prjxray.segmaker import Segmaker
from prjxray.bitfilter import get_bitfilter
import argparse

parser = argparse.ArgumentParser(description="Generate int segfiles")
parser.add_argument(
    '--todo', action='store', default='../../todo.txt', help='todo file path')
parser.add_argument(
    '--design',
    action='store',
    default='design.txt',
    help='design description file path')
parser.add_argument('--verbose', action='store_true', help='')
parser.add_argument(
    '--bits', action='store', default='design.bits', help='bits file path')

args = parser.parse_args()
segmk = Segmaker(args.bits)

verbose = args.verbose

tiledata = dict()
pipdata = dict()
ignpip = set()
todo = set()

print("Loading todo from %s." % args.todo)
with open(args.todo, "r") as f:
    for line in f:
        line = tuple(line.strip().split("."))
        verbose and print('todo', line)
        todo.add(line)

print("Loading tags from %s." % args.design)
with open(args.design, "r") as f:
    for line in f:
        tile, pip, src, dst, pnum, pdir = line.split()

        if not tile.startswith('INT_'):
            continue

        _, pip = pip.split(".")
        _, src = src.split("/")
        _, dst = dst.split("/")
        pnum = int(pnum)
        pdir = int(pdir)

        if tile not in tiledata:
            tiledata[tile] = {"pips": set(), "srcs": set(), "dsts": set()}

        tile_lr = ("_".join(tile.split("_")[0:2]))
        if (tile_lr, pip) in pipdata:
            assert pipdata[(tile_lr, pip)] == (src, dst)
        else:
            pipdata[(tile_lr, pip)] = (src, dst)

        tiledata[tile]["pips"].add(pip)
        tiledata[tile]["srcs"].add(src)
        tiledata[tile]["dsts"].add(dst)

        if pdir == 0:
            tiledata[tile]["srcs"].add(dst)
            tiledata[tile]["dsts"].add(src)

        t = ("_".join(tile.split("_")[0:2]), dst, src)
        if pnum == 1 or pdir == 0:
            verbose and print('ignore pnum == 1 or pdir == 0: ', pip)
            ignpip.add(t)

        if t not in todo:
            verbose and print('ignore not todo: ', t)
            ignpip.add(t)

for tile, pips_srcs_dsts in tiledata.items():
    pips = pips_srcs_dsts["pips"]
    srcs = pips_srcs_dsts["srcs"]
    dsts = pips_srcs_dsts["dsts"]
    for (tile_lr, pip), src_dst in pipdata.items():
        src, dst = src_dst
        t = (tile_lr, dst, src)
        if t in ignpip:
            pass
        elif pip in pips:
            segmk.add_tile_tag(tile, "%s.%s" % (dst, src), 1)
        elif src_dst[1] not in dsts:
            segmk.add_tile_tag(tile, "%s.%s" % (dst, src), 0)

segmk.compile(bitfilter=get_bitfilter(os.getenv('XRAY_DATABASE'), 'INT'))
segmk.write()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

from prjxray.segmaker import Segmaker

verbose = False

segmk = Segmaker("design.bits")

# Can fit 4 per CLB
# BELable
multi_bels_by = [
    'SRL16E',
    'SRLC32E',
]
# Not BELable
multi_bels_bn = [
    'RAM32X1S',
    'RAM64X1S',
]


def load_tcl():
    f = open('design.csv', 'r')
    f.readline()
    ret = {}
    for l in f:
        l = l.strip()
        tile, site, bel, cell, ref_name, prim_type = l.split(',')
        ret[bel] = ref_name
    return ret


design = load_tcl()

print("Loading tags")
'''
module,loc,bela,belb,belc,beld
my_ram_N,SLICE_X12Y100,SRLC32E,SRL16E,SRLC32E,LUT6
my_ram_N,SLICE_X12Y101,SRLC32E,SRLC32E,SRLC32E,SRLC32E
my_RAM256X1S,SLICE_X12Y102,None,0,,
'''
f = open('params.csv', 'r')
f.readline()
for l in f:
    l = l.strip()
    module, loc, p0, p1, p2, p3 = l.split(',')

    segmk.add_site_tag(
        loc, "WA7USED",
        module in ('my_RAM128X1D', 'my_RAM128X1S_N', 'my_RAM256X1S'))
    segmk.add_site_tag(loc, "WA8USED", module == 'my_RAM256X1S')

    bels_tcl = [design.get("%s/%c6LUT" % (loc, bel), None) for bel in "ABCD"]

    # (a, b, c, d)
    # Size set for RAM32X1S, RAM32X1D, and SRL16E
    size = [0, 0, 0, 0]
    # SRL set for SRL* primitives
    srl = [0, 0, 0, 0]
    # RAM set for RAM* primitives
    ram = [0, 0, 0, 0]

    verbose and print('%s' % loc)
    verbose and print('  %s %s %s %s' % tuple(bels_tcl))

    if module == 'my_ram_N':
        # Each one of: SRL16E, SRLC32E, LUT6
        bels = [p0, p1, p2, p3]
        verbose and print('  %s %s %s %s' % tuple(bels))
        assert bels == bels_tcl

        # Clock Enable (CE) clock gate only enabled if we have clocked elements
        # A pure LUT6 does not, but everything else should
        segmk.add_site_tag(
            loc, "WEMUX.CE", bels != ['LUT6', 'LUT6', 'LUT6', 'LUT6'])

        beli = 0
        for which, bel in zip('ABCD', bels):
            if bel == 'SRL16E':
                size[beli] = 1
            if bel in ('SRL16E', 'SRLC32E'):
                srl[beli] = 1
            beli += 1
    else:
        n = p0
        if n:
            n = int(n)
        # Unused. Just to un-alias mux
        #_ff = int(p1)

        # Can pack 4 into a CLB
        # D is always occupied first (due to WA/A sharing on D)
        # TODO: maybe investigate ROM primitive for completeness
        pack4 = [
            # (a, b, c, d)
            (0, 0, 0, 1),
            (1, 0, 0, 1),
            (1, 1, 0, 1),
            (1, 1, 1, 1),
        ]
        # Uses CD first
        pack2 = [
            (0, 0, 1, 1),
            (1, 1, 1, 1),
        ]
        has_bel_tcl = tuple([int(bool(x)) for x in bels_tcl])

        # Always use all 4 sites
        if module in ('my_RAM32M', 'my_RAM64M', 'my_RAM128X1D',
                      'my_RAM256X1S'):
            ram = (1, 1, 1, 1)
        # Only can occupy CD I guess
        elif module == 'my_RAM32X1D':
            ram = (0, 0, 1, 1)
        # Uses 2 sites at a time
        elif module in ('my_RAM64X1D_N', 'my_RAM128X1S_N'):
            ram = pack2[n - 1]
        # Uses 1 site at a time
        elif module in ('my_RAM32X1S_N', 'my_RAM64X1S_N'):
            ram = pack4[n - 1]
        else:
            assert (0)
        verbose and print('  %s %s %s %s' % tuple(ram))
        verbose and print('  %s %s %s %s' % tuple(has_bel_tcl))
        # assert ram == ram_tcl
        # Hack: reject if something unexpected got packed in
        # TODO: place dummy LUTs to exclude placement?
        if ram != has_bel_tcl:
            continue

        # All entries here require D
        assert (ram[3])

        if module == 'my_RAM32X1D':
            # Occupies CD
            size[2] = 1
            size[3] = 1
        elif module == 'my_RAM32M':
            size = [1, 1, 1, 1]
        elif module == 'my_RAM32X1S_N':
            size = pack4[n - 1]
            if size != has_bel_tcl:
                continue
        else:
            assert (not module.startswith('my_RAM32'))

    # Now commit bits after marking 1's
    for beli, bel in enumerate('ABCD'):
        segmk.add_site_tag(loc, "%sLUT.RAM" % bel, ram[beli])
        # FIXME: quick fix
        segmk.add_site_tag(loc, "%sLUT.SRL" % bel, srl[beli])
        segmk.add_site_tag(loc, "%sLUT.SMALL" % bel, size[beli])


def bitfilter(frame_idx, bit_idx):
    # Hack to remove aliased PIP bits on CE
    # We should either mix up routing more or exclude previous DB entries
    return (frame_idx, bit_idx) not in [(0, 27), (1, 25), (1, 26), (1, 29)]


segmk.compile(bitfilter=bitfilter)
segmk.write()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import sys
import re

from prjxray.segmaker import Segmaker

tags = dict()
en_tags = dict()

print("Preload all tags.")
for arg in sys.argv[1:]:
    with open(arg + ".txt", "r") as f:
        for line in f:
            tile, pip = line.split()
            _, pip = pip.split("/")
            tile_type, pip = pip.split(".")
            src, dst = pip.split("->>")
            tag = "%s.%s" % (dst, src)
            tags[tag] = dst
            if "HCLK_CK_BUFH" in src:
                en_tag = "ENABLE_BUFFER.%s" % src
                en_tags[en_tag] = src

for arg in sys.argv[1:]:
    print("Processing %s." % arg)
    segmk = Segmaker(arg + ".bits")

    tiledata = dict()
    pipdata = dict()
    ignpip = set()

    with open(arg + ".txt", "r") as f:
        for line in f:
            tile, pip = line.split()
            _, pip = pip.split("/")
            tile_type, pip = pip.split(".")
            src, dst = pip.split("->>")

            # FIXME: workaround for https://github.com/SymbiFlow/prjxray/issues/392
            if "CLB_IO_CLK" not in segmk.grid[tile]["bits"]:
                print("WARNING: dropping tile %s" % tile)
                continue
            tag = "%s.%s" % (dst, src)

            def add_tile_tag(tag, val):
                # Workaround for https://github.com/SymbiFlow/prjxray/issues/396
                # TODO: drop from tcl or make an ignpip
                if re.match(r"HCLK_CK_INOUT_.*", tag):
                    # print("Dropping %s %s" % (tag, val))
                    return
                segmk.add_tile_tag(tile, tag, val)

            add_tile_tag(tag, 1)
            if "HCLK_CK_BUFH" in src:
                en_tag = "ENABLE_BUFFER.%s" % src
                add_tile_tag(en_tag, 1)
            for tag, tag_dst in tags.items():
                if tag_dst != dst:
                    add_tile_tag(tag, 0)
            for en_tag, en_tag_src in en_tags.items():
                if en_tag_src != src:
                    add_tile_tag(en_tag, 0)

    segmk.compile()
    segmk.write(arg)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
"""
Add bits that are considered always on to the db file.

This script is Zynq specific.

There are three bits that are present in all Zynq bitstreams.
The investigation that was done to reach this conclusion is captured on GH
(https://github.com/SymbiFlow/prjxray/issues/746)
In brief, these bits seem to be bitstream properties related,
but no evidence of this could be found.
Due to the fact that the base address of these bits is the same as for the
CFG_CENTER_MID tile it has been decided to append the bits to its db file.
"""

import sys

constant_bits = {
    "CFG_CENTER_MID.ALWAYS_ON_PROP1": "26_2206",
    "CFG_CENTER_MID.ALWAYS_ON_PROP2": "26_2207",
    "CFG_CENTER_MID.ALWAYS_ON_PROP3": "27_2205"
}

with open(sys.argv[1], "a") as f:
    for bit_name, bit_value in constant_bits.items():
        f.write(bit_name + " " + bit_value + "\n")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import json

from prjxray.segmaker import Segmaker


def bitfilter(frame, bit):
    return True


def main():
    segmk = Segmaker("design.bits")

    print("Loading tags")
    with open('params.json') as f:
        params = json.load(f)

    for row in params:
        base_name = 'BUFR_Y{}'.format(row['y'])

        segmk.add_tile_tag(
            row['tile'], '{}.IN_USE'.format(base_name), row['IN_USE'])

        if not row['IN_USE']:
            continue

        segmk.add_tile_tag(
            row['tile'], '{}.BUFR_DIVIDE.BYPASS'.format(base_name),
            '"BYPASS"' == row['BUFR_DIVIDE'])
        for opt in range(1, 9):
            if row['BUFR_DIVIDE'] == str(opt):
                segmk.add_tile_tag(
                    row['tile'], '{}.BUFR_DIVIDE.D{}'.format(base_name, opt),
                    1)
            elif '"BYPASS"' == row['BUFR_DIVIDE']:
                segmk.add_tile_tag(
                    row['tile'], '{}.BUFR_DIVIDE.D{}'.format(base_name, opt),
                    0)

    segmk.compile(bitfilter=bitfilter)
    segmk.write()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import json
import io
import os
import random
random.seed(int(os.getenv("SEED"), 16))
from prjxray import util
from prjxray import lut_maker
from prjxray import verilog
from prjxray.db import Database


def gen_sites():
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()
    for tile_name in sorted(grid.tiles()):
        if not 'CFG_CENTER_MID' in tile_name:
            continue
        print("// tile: " + str(tile_name))
        loc = grid.loc_of_tilename(tile_name)
        gridinfo = grid.gridinfo_at_loc(loc)

        sites = {}
        print("// " + str(gridinfo.sites.items()))
        for site_name, site_type in gridinfo.sites.items():
            if site_type == 'STARTUP':
                print("// got site: " + str(site_name))
                sites[site_type] = site_name

        if sites:
            yield tile_name, sites

def run():
    params = {
        "tiles": [],
    }

    for tile, sites in gen_sites():
        for site_type, site in sites.items():
            p = {}
            p['tile'] = tile
            p['site'] = site

            p['CONNECTION'] = random.choice(
                (
                    'HARD_ZERO',
                    # hard zero or hard one does not make a difference
                    # it only seems to matter if it is connected to a clock net or not
                    #'HARD_ONE',
                    'CLOCK',
                ))

            params['tiles'].append(p)

    print(
    '''
module top (input  wire clk);
    (* KEEP, DONT_TOUCH *)
    STARTUPE2 STARTUPE2 (
        .CLK(1'b0),
        .GSR(1'b0),
        .GTS(1'b0),
        .KEYCLEARB(1'b1),
        .PACK(1'b0),
        .PREQ(),
  
        // Drive clock.''')

    connection = p['CONNECTION']

    if connection == "HARD_ZERO":
        print("        .USRCCLKO (1'b0),")
    elif connection == "HARD_ONE":
        print("        .USRCCLKO (1'b1),")
    else:    
        print("        .USRCCLKO (clk),")

    print(
    '''
        .USRCCLKTS(1'b0),  
        .USRDONEO (1'b0),
        .USRDONETS(1'b1),
        .CFGCLK(),
        .CFGMCLK(),
        .EOS()
    );

endmodule
''')

    with open('params.json', 'w') as f:
        json.dump(params, f, indent=2)

if __name__ == '__main__':
    run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import argparse
import subprocess
import tempfile
import os.path
import glob


def main():
    parser = argparse.ArgumentParser(
        description=
        "Tool for comparing database JSON outputs from two database's.")
    parser.add_argument('a_db')
    parser.add_argument('b_db')

    args = parser.parse_args()

    assert os.path.isdir(args.a_db)
    assert os.path.isdir(args.b_db)

    for a_json_in in glob.glob('{}/*.json'.format(args.a_db)):
        a_json_base = os.path.basename(a_json_in)
        b_json_in = '{}/{}'.format(args.b_db, a_json_base)

        if not os.path.exists(b_json_in):
            print('{} not found!'.format(b_json_in))
            continue

        with tempfile.NamedTemporaryFile(suffix="_a_{}".format(
                a_json_base)) as a_json_out, tempfile.NamedTemporaryFile(
                    suffix="_b_{}".format(a_json_base)) as b_json_out:
            subprocess.check_call(
                "python3 -m utils.xjson {}".format(a_json_in),
                shell=True,
                stdout=a_json_out)
            subprocess.check_call(
                "python3 -m utils.xjson {}".format(b_json_in),
                shell=True,
                stdout=b_json_out)

            print("Comparing {}".format(a_json_base))
            subprocess.call(
                "diff -U 10 {} {}".format(a_json_out.name, b_json_out.name),
                shell=True)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2021  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import argparse
import multiprocessing as mp
import os
import re
import subprocess
import yaml

from prjxray import util


def worker(arglist):
    part, cwd = arglist
    cmd = "make roi_only"

    env = os.environ.copy()
    env['XRAY_PART'] = part

    print("running subprocess")
    subprocess.run(cmd.split(' '), check=True, env=env, cwd=cwd)


def main():
    """Rois all parts for a family by calling "make roi_only" over all parts
    with the same device as XRAY_PART.

    Example:
        prjxray$ ./utils/roi_all.py --db-root database/artix7/ --part xc7a100tfgg676-1
    """
    env = os.environ.copy()
    db_root = util.get_db_root()
    assert db_root
    part = util.get_part()
    assert part

    information = util.get_part_information(db_root, part)

    valid_devices = []
    for name, device in util.get_devices(db_root).items():
        if device['fabric'] == information['device']:
            valid_devices.append(name)

    tasks = []

    for part, data in util.get_parts(db_root).items():
        if data['device'] in valid_devices:
            cwd = os.getenv('XRAY_FUZZERS_DIR')

            tasks.append((part, cwd))

    with mp.Pool() as pool:
        for _ in pool.imap_unordered(worker, tasks):
            pass


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2023  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import json

from prjxray.segmaker import Segmaker, add_site_group_zero
from prjxray import verilog


def bitfilter(frame, word):
    return frame >= 27

def main():
    segmk = Segmaker("design.bits", verbose=True)

    # Load tags
    with open("params.json", "r") as fp:
        data = json.load(fp)

    odelay_types = ["FIXED", "VARIABLE", "VAR_LOAD"]
    delay_srcs = ["ODATAIN"]

    # Output tags
    for params in data:
        if params['ODELAY_BYPASS']:
            prims = params['ODELAY_NOT_IN_USE'].split(" ")
            segmk.add_site_tag(prims[0], 'IN_USE', False)
            segmk.add_site_tag(prims[1], 'IN_USE', False)
            continue
        segmk.add_site_tag(params['ODELAY_IN_USE'], 'IN_USE', True)
        segmk.add_site_tag(params['ODELAY_NOT_IN_USE'], 'IN_USE', False)

        loc = verilog.unquote(params["LOC"])

        # Delay type
		# VAR_LOAD and VAR_LOAD_PIPE are the same
        value = verilog.unquote(params["ODELAY_TYPE"])
        add_site_group_zero(
            segmk, loc, "ODELAY_TYPE_", odelay_types, "FIXED", value)

        # Delay value
        value = int(params["ODELAY_VALUE"])
        for i in range(5):
            segmk.add_site_tag(
                loc, "ODELAY_VALUE[%01d]" % i, ((value >> i) & 1) != 0)
            segmk.add_site_tag(
                loc, "ZODELAY_VALUE[%01d]" % i, ((value >> i) & 1) == 0)

        value = verilog.unquote(params["CINVCTRL_SEL"])
        segmk.add_site_tag(loc, "CINVCTRL_SEL", int(value == "TRUE"))

        value = verilog.unquote(params["PIPE_SEL"])
        segmk.add_site_tag(loc, "PIPE_SEL", int(value == "TRUE"))

        if "IS_C_INVERTED" in params and verilog.unquote(params["CINVCTRL_SEL"]) != "TRUE":
            segmk.add_site_tag(
                loc, "IS_C_INVERTED", int(params["IS_C_INVERTED"]))
            segmk.add_site_tag(loc, "ZINV_C", 1 ^ int(params["IS_C_INVERTED"]))

        value = verilog.unquote(params["HIGH_PERFORMANCE_MODE"])
        segmk.add_site_tag(
            loc, "HIGH_PERFORMANCE_MODE", int(value == "TRUE"))

        segmk.add_site_tag(
            loc, "ZINV_ODATAIN", 1 ^ int(params["IS_ODATAIN_INVERTED"]))

    segmk.compile(bitfilter=bitfilter)
    segmk.write()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

from utils import xjson


def load_tiles(tiles_fn):
    '''
    "$type $tile $grid_x $grid_y $skip_tile $clock_region $typed_sites"
    typed_sites: foreach t $site_types s $sites
    '''
    tiles = list()

    with open(tiles_fn) as f:
        for line in f:
            # CLBLM_L CLBLM_L_X10Y98 30 106 SLICEL SLICE_X13Y98 SLICEM SLICE_X12Y98
            record = line.split()
            tile_type, tile_name, grid_x, grid_y, skip_tile = record[0:5]
            grid_x, grid_y = int(grid_x), int(grid_y)
            skip_tile = int(skip_tile) != 0
            sites = {}

            # prohibits is the list of sites that the Vivado placer will not
            # place at.
            #
            # Speculation: These sites are prohibited not because the hardware
            # doesn't work, but because the interconnect around these sites is
            # extremely narrow due to the hardblocks to the left and right of
            # tiles.  As a result, these sites should be avoided because
            # congestion and delays when using these sites might be very very
            # high.
            prohibits = []
            clock_region = None
            if len(record) >= 6:
                clock_region = record[5]
                if clock_region == "NA":
                    clock_region = None
                for i in range(6, len(record), 3):
                    site_type, site_name, prohibited = record[i:i + 3]
                    sites[site_name] = site_type
                    if int(prohibited):
                        prohibits.append(site_name)

            if not skip_tile:
                tile = {
                    'type': tile_type,
                    'name': tile_name,
                    'grid_x': grid_x,
                    'grid_y': grid_y,
                    'sites': sites,
                    'prohibited_sites': sorted(prohibits),
                    'clock_region': clock_region,
                }
            else:
                # Replace tiles within the exclude_roi with NULL tiles to
                # ensure no gaps in the tilegrid.
                #
                # The name will reflect the original tile.
                tile = {
                    'type': 'NULL',
                    'name': tile_name,
                    'grid_x': grid_x,
                    'grid_y': grid_y,
                    'sites': {},
                    'prohibited_sites': [],
                    'clock_region': clock_region,
                }

            tiles.append(tile)

    return tiles


def load_pin_functions(pin_func_fn):
    pin_functions = {}

    with open(pin_func_fn) as f:
        for line in f:
            site, pin_func = line.split()
            assert site not in pin_functions, site
            pin_functions[site] = pin_func

    return pin_functions


def make_database(tiles, pin_func):
    # tile database with X, Y, and list of sites
    # tile name as keys
    database = dict()

    for tile in tiles:
        database[tile["name"]] = {
            "type": tile["type"],
            "sites": tile["sites"],
            "grid_x": tile["grid_x"],
            "grid_y": tile["grid_y"],
            "bits": {},
            "pin_functions": {},
            "prohibited_sites": tile['prohibited_sites'],
        }

        if tile["clock_region"]:
            database[tile["name"]]["clock_region"] = tile["clock_region"]

        for site in database[tile["name"]]["sites"]:
            if site in pin_func:
                database[tile["name"]]["pin_functions"][site] = pin_func[site]

    return database


def run(tiles_fn, pin_func_fn, json_fn, verbose=False):
    # Load input files
    tiles = load_tiles(tiles_fn)

    # Read site map
    pin_func = load_pin_functions(pin_func_fn)

    # Index input
    database = make_database(tiles, pin_func)

    # Save
    xjson.pprint(open(json_fn, 'w'), database)


def main():
    import argparse

    parser = argparse.ArgumentParser(
        description='Generate tilegrid.json from bitstream deltas')

    parser.add_argument('--verbose', action='store_true', help='')
    parser.add_argument('--out', default='/dev/stdout', help='Output JSON')
    parser.add_argument(
        '--tiles',
        default='tiles.txt',
        help='Input tiles.txt tcl output',
        required=True)
    parser.add_argument(
        '--pin_func', help='List of sites with pin functions', required=True)
    args = parser.parse_args()

    run(args.tiles, args.pin_func, args.out, verbose=args.verbose)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
"""
This script loads the pin dump of the desired BEL from Vivado and groups pins into ports.

Ports that are not connected to the PL are not included. These ports are usually test and
debug related ports, which are not useful from a P&R perspective.

Ports are then written to a JSON file.
"""
import argparse
import csv
import json
import re

from collections import defaultdict

from prjxray.util import OpenSafeFile


def main():

    BUS_REGEX = re.compile("(.*[A-Z_])([0-9]+)$")

    # Parse arguments
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument("csv", type=str, help="BEL pin dump file")
    parser.add_argument(
        "json",
        type=str,
        help="Output JSON file with BEL pins grouped into ports")

    parser.add_argument(
        "--special-pins",
        default="",
        type=str,
        required=False,
        help="Some pins cannot be decoded with the regex")

    args = parser.parse_args()

    # Load pin dump
    with OpenSafeFile(args.csv, "r") as fp:
        pin_dump = list(csv.DictReader(fp))

    # Group pins into ports
    ports = defaultdict(lambda: {"direction": None, "width": 0})

    special_pins = args.special_pins.split(",")

    for pin in list(pin_dump):
        pin_name = pin["name"]

        name = None
        if pin_name in special_pins:
            # Full match
            name = pin_name
        else:
            # Partial match
            for special_pin in special_pins:
                if pin_name.startswith(special_pin):
                    name = special_pin
                    break

        if name is None:
            match = BUS_REGEX.match(pin_name)
            if match:
                name = match.group(1)
            else:
                name = pin_name

        # Get direction
        is_input = int(pin["is_input"])
        is_output = int(pin["is_output"])
        is_clock = int(pin["is_clock"])

        if is_input:
            direction = "input"
            if is_clock:
                direction = "clock"
        elif is_output:
            direction = "output"
        else:
            assert False, pin

        # Add to port
        port = ports[name]

        if port["direction"] is None:
            port["direction"] = direction
        else:
            assert port["direction"] == direction, (port, direction, name)

        port["width"] += 1

    # Write pin ports to a JSON file
    with OpenSafeFile(args.json, "w") as fp:
        json.dump(ports, fp, indent=1, sort_keys=True)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
""" Math models are used to represent abstract operations for the timing models.
This is useful for creating excel workbooks that can update dynamically, or
generating a model with symbolic constants to be backsolved.
"""


class ExcelMathModel(object):
    """ Math model used for outputting to an dyanmic Excel sheet. """

    def __init__(self):
        pass

    def sum(self, elems):
        sum_val = '(' + ' + '.join(elems) + ')'
        if sum_val == '()':
            return '0'
        else:
            return sum_val

    def product(self, elems):
        sum_val = '(' + ' * '.join(elems) + ')'
        if sum_val == '()':
            return '1'
        else:
            return sum_val

    def plus(self, a, b):
        return self.sum((a, b))

    def divide(self, a, b):
        return '({}/{})'.format(a, b)

    def multiply(self, a, b):
        return '({}*{})'.format(a, b)

    def eval(self, elem):
        return '=' + elem


def PythonMathModel(object):
    """ Math model used for outputting values equalated immediately. """

    def __init__(self):
        pass

    def sum(self, elems):
        return sum(elems)

    def product(self, elems):
        v = 1.0
        for elem in elems:
            v *= elem
        return v

    def divide(self, a, b):
        return a / b

    def plus(self, a, b):
        return a + b

    def multiply(self, a, b):
        return a * b

    def eval(self, elem):
        return elem



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import os
import random
import math
random.seed(int(os.getenv("SEED"), 16))
from prjxray import util
from prjxray.lut_maker import LutMaker
from prjxray.db import Database

from ports import ports


def print_site(ports, luts, site, site_type):
    verilog_ports = ""
    verilog_wires = ""

    for port, width in ports:
        verilog_ports += """
    .{port}({port}_{site}),""".format(
            port=port, site=site)
        verilog_wires += "wire [{}:0] {}_{};\n".format(width - 1, port, site)

        for idx in range(0, width):
            verilog_wires += "assign {}_{}[{}] = {};\n".format(
                port, site, idx, luts.get_next_output_net())

        verilog_wires += "\n"

    verilog_ports = verilog_ports.rstrip(",")

    print(
        """
{wires}

(* KEEP, DONT_TOUCH, LOC = "{site}" *)
{site_type} {site}_instance (
{ports}
);""".format(
            wires=verilog_wires,
            ports=verilog_ports,
            site=site,
            site_type=site_type))


def main():
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()

    luts = LutMaker()

    def gen_sites(desired_site_type):
        for tile_name in sorted(grid.tiles()):
            loc = grid.loc_of_tilename(tile_name)
            gridinfo = grid.gridinfo_at_loc(loc)
            for site, site_type in gridinfo.sites.items():
                if site_type == desired_site_type:
                    yield tile_name, site

    print('''
module top();

    (* KEEP, DONT_TOUCH *)
    LUT6 dummy();
''')

    for site_type in ["GTPE2_CHANNEL", "GTPE2_COMMON"]:
        for _, site in gen_sites(site_type):
            print_site(ports[site_type], luts, site, site_type)

    for l in luts.create_wires_and_luts():
        print(l)

    print('endmodule')


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

# https://symbiflow.github.io/prjxray-db/
# https://symbiflow.github.io/prjxray-db/artix7/

import os, sys, json, re
from io import StringIO
from prjxray.util import get_fabric_for_part


def mk_get_setting(settings_filename):
    if settings_filename:
        settings = {}
        with open(settings_filename) as f:
            for line in f:
                line = line.strip()
                if not line.startswith("export "):
                    continue
                key, value = line[7:].split('=', 1)
                settings[key] = value[1:-1]

        assert len(settings), (settings_filename, settings)
        assert settings['XRAY_DATABASE'], pprint.pformat(settings)
        settings['XRAY_DATABASE_DIR'] = os.path.abspath(
            os.path.join(
                os.path.dirname(settings_filename),
                '..',
                'database',
            ), )
        return lambda name: settings[name]
    else:
        return os.getenv


get_setting = mk_get_setting(None)


def db_open(fn, db_dir):
    filename = os.path.join(db_dir, fn)
    if not os.path.exists(filename):
        return StringIO("")
    return open(os.path.join(db_dir, fn))


def out_open(fn, output):
    os.makedirs(output, exist_ok=True)
    fp = os.path.join(output, fn)
    print("Writing %s" % fp)
    return open(fp, "w")


class UnionFind:
    def __init__(self):
        self.parents = dict()

    def make(self, value):
        if value not in self.parents:
            self.parents[value] = value

    def find(self, value):
        self.make(value)
        if self.parents[value] != value:
            retval = self.find(self.parents[value])
            self.parents[value] = retval
        return self.parents[value]

    def union(self, v1, v2):
        a = self.find(v1)
        b = self.find(v2)
        if a != b:
            self.parents[a] = b


def db_read(dbstate, tiletype, db_dir):
    dbstate.cfgbits[tiletype] = dict()
    dbstate.cfgbits_r[tiletype] = dict()
    dbstate.maskbits[tiletype] = set()
    dbstate.ppips[tiletype] = dict()
    dbstate.routebits[tiletype] = dict()
    dbstate.routezbits[tiletype] = dict()

    def add_pip_bits(tag, bits):
        if tag not in dbstate.routebits[tiletype]:
            dbstate.routebits[tiletype][tag] = set()
            dbstate.routezbits[tiletype][tag] = set()
        for bit in bits:
            if bit[0] == "!":
                if bit[1:] not in dbstate.routezbits[tiletype]:
                    dbstate.routezbits[tiletype][bit[1:]] = set()
                dbstate.routezbits[tiletype][bit[1:]].add(tag)
            else:
                if bit not in dbstate.routebits[tiletype]:
                    dbstate.routebits[tiletype][bit] = set()
                dbstate.routebits[tiletype][bit].add(tag)

    def add_cfg_bits(tag, bits):
        if tag not in dbstate.cfgbits[tiletype]:
            dbstate.cfgbits[tiletype][tag] = set()
        for bit in bits:
            dbstate.cfgbits[tiletype][tag].add(bit)
            if bit not in dbstate.cfgbits_r[tiletype]:
                dbstate.cfgbits_r[tiletype][bit] = set()
            dbstate.cfgbits_r[tiletype][bit].add(tag)

    with db_open("segbits_%s.db" % tiletype, db_dir) as f:
        for line in f:
            line = line.split()
            tag, bits = line[0], line[1:]

            if tiletype in ["int_l", "int_r", "hclk_l", "hclk_r"]:
                add_pip_bits(tag, bits)

            elif tiletype in ["clbll_l", "clbll_r", "clblm_l", "clblm_r"] and \
                    re.search(r"(\.[ABCD].*MUX\.)|(\.PRECYINIT\.)", tag):
                add_pip_bits(tag, bits)

            else:
                add_cfg_bits(tag, bits)

    with db_open("ppips_%s.db" % tiletype, db_dir) as f:
        for line in f:
            tag, typ = line.split()
            dbstate.ppips[tiletype][tag] = typ

    if tiletype not in ["int_l", "int_r"]:
        with db_open("mask_%s.db" % tiletype, db_dir) as f:
            for line in f:
                tag, bit = line.split()
                assert tag == "bit"
                dbstate.maskbits[tiletype].add(bit)
    else:
        for t in ["clbll_l", "clbll_r", "clblm_l", "clblm_r", "dsp_l", "dsp_r",
                  "bram_l", "bram_r"]:
            with db_open("mask_%s.db" % t, db_dir) as f:
                for line in f:
                    tag, bit = line.split()
                    assert tag == "bit"
                    frameidx, bitidx = bit.split("_")
                    dbstate.maskbits[tiletype].add(
                        "%02d_%02d" % (int(frameidx), int(bitidx) % 64))


def init_bitdb():
    clb_bitgroups_db = [
        # copy&paste from zero_db in dbfixup.py
        "00_21 00_22 00_26 01_28|00_25 01_20 01_21 01_24",
        "00_23 00_30 01_22 01_25|00_27 00_29 01_26 01_29",
        "01_12 01_14 01_16 01_18|00_10 00_11 01_09 01_10",
        "00_13 01_17 00_15 00_17|00_18 00_19 01_13 00_14",
        "00_34 00_38 01_33 01_37|00_35 00_39 01_38 01_40",
        "00_33 00_41 01_32 01_34|00_37 00_42 01_36 01_41",

        # other manual groupings for individual bits
        "00_02 00_05 00_09 01_04|00_07 01_05 01_06",
        "00_01 00_06 01_00 01_08|00_03 01_01 01_02",
        "00_59 01_54 01_58 01_61|00_57 00_58 01_56",
        "00_55 00_63 01_57 01_62|00_61 00_62 01_60",
        "00_43 00_47 00_50 00_53 00_54 01_42|00_51 01_50 01_52",
        "00_49 01_44 01_45 01_48 01_49 01_53|00_45 00_46 01_46",
    ]

    hclk_bitgroups_db = [
        # manual groupings
        "03_14 03_15 04_14 04_15|00_15 00_16 01_14 01_15",
        "02_16 03_16 04_16 05_16|02_14 02_15 05_14 05_15",
        "02_18 02_19 05_18 05_19|00_17 00_18 01_16 01_17",
        "03_18 03_19 04_18 04_19|02_17 03_17 04_17 05_17",
        "02_20 02_21 05_20 05_21|02_22 03_22 04_22 05_22",
        "02_29 03_29 04_29 05_29|03_30 03_31 04_30 04_31",
        "02_26 02_27 05_26 05_27|02_28 03_28 04_28 05_28",
        "02_23 03_23 04_23 05_23|03_24 03_25 04_24 04_25",
    ]

    # groupings for SNWE bits in frames 2..7
    for i in range(0, 64, 4):
        clb_bitgroups_db.append(
            "02_%02d 03_%02d 05_%02d 06_%02d 07_%02d|05_%02d 03_%02d 04_%02d 04_%02d"
            % (i + 1, i, i, i, i + 1, i + 3, i + 1, i + 1, i + 2))
        clb_bitgroups_db.append(
            "02_%02d 04_%02d 05_%02d 05_%02d 06_%02d|02_%02d 03_%02d 04_%02d 07_%02d"
            % (i + 2, i, i + 1, i + 2, i + 2, i + 3, i + 2, i + 3, i + 3))

    return clb_bitgroups_db, hclk_bitgroups_db


def init_hclk_bits(hclk_bitgroups_db):
    hclk_left_bits = set()
    hclk_right_bits = set()

    for entry in hclk_bitgroups_db:
        a, b = entry.split("|")
        for bit in a.split():
            hclk_left_bits.add(bit)
        for bit in b.split():
            hclk_right_bits.add(bit)

    return hclk_left_bits, hclk_right_bits


def init_clb_bits(clb_bitgroups_db):
    clb_left_bits = set()
    clb_right_bits = set()

    for entry in clb_bitgroups_db:
        a, b = entry.split("|")
        for bit in a.split():
            clb_left_bits.add(bit)
        for bit in b.split():
            clb_right_bits.add(bit)

    return clb_left_bits, clb_right_bits


class DBState():
    def __init__(self):
        self.cfgbits = dict()
        self.cfgbits_r = dict()
        self.maskbits = dict()
        self.ppips = dict()
        self.routebits = dict()
        self.routezbits = dict()


class Tweaks():
    def __init__(self):
        pass


def load_tilegrid(db_dir, fabric, verbose=False, allow_fake=False):
    print("Loading tilegrid.")
    with db_open(os.path.join(fabric, "tilegrid.json"), db_dir) as f:
        data = f.read()
        if not data:
            assert allow_fake, 'No tilegrid.json found'
            print('WARNING: loading fake tilegrid')
            grid = {
                "NULL": {
                    "grid_x": 0,
                    "grid_y": 0,
                    "type": "NULL",
                }
            }
        else:
            grid = json.loads(data)
    return grid


def db_reads(dbstate, db_dir):

    db_read(dbstate, "int_l", db_dir)
    db_read(dbstate, "int_r", db_dir)

    db_read(dbstate, "hclk_l", db_dir)
    db_read(dbstate, "hclk_r", db_dir)

    db_read(dbstate, "clbll_l", db_dir)
    db_read(dbstate, "clbll_r", db_dir)

    db_read(dbstate, "clblm_l", db_dir)
    db_read(dbstate, "clblm_r", db_dir)

    db_read(dbstate, "dsp_l", db_dir)
    db_read(dbstate, "dsp_r", db_dir)

    db_read(dbstate, "bram_l", db_dir)
    db_read(dbstate, "bram_r", db_dir)


def place_tiles(grid):
    grid_map = dict()
    grid_range = None

    for tilename, tiledata in grid.items():
        grid_x = tiledata["grid_x"]
        grid_y = tiledata["grid_y"]
        grid_map[(grid_x, grid_y)] = tilename

        if grid_range is None:
            grid_range = [grid_x, grid_y, grid_x, grid_y]
        else:
            grid_range[0] = min(grid_range[0], grid_x)
            grid_range[1] = min(grid_range[1], grid_y)
            grid_range[2] = max(grid_range[2], grid_x)
            grid_range[3] = max(grid_range[3], grid_y)
    return grid_map, grid_range


def tile_bgcolor(tiledata):
    bgcolor = "#eeeeee"

    # INT - Blue
    if tiledata["type"] in ["INT_L", "INT_R"]: bgcolor = "#aaaaff"
    elif "INT_FEEDTHRU" in tiledata["type"]: bgcolor = "#ddddff"

    # CLBL - Yellow
    if tiledata["type"] in ["CLBLL_L", "CLBLL_R"]: bgcolor = "#ffffaa"
    # CLBM - Red
    if tiledata["type"] in ["CLBLM_L", "CLBLM_R"]: bgcolor = "#ffaaaa"

    # CLK - Green
    if tiledata["type"] in ["HCLK_L", "HCLK_R"]: bgcolor = "#aaffaa"
    elif "CLK" in tiledata["type"]: bgcolor = "#66ff66"
    elif "CMT" in tiledata["type"]: bgcolor = "#22ff22"

    # BRAM - Cyan
    if tiledata["type"] in ["BRAM_INT_INTERFACE_L", "BRAM_L"]:
        bgcolor = "#aaffff"
    if tiledata["type"] in ["BRAM_INT_INTERFACE_R", "BRAM_R"]:
        bgcolor = "#aaffff"

    # DSP - Purple
    if tiledata["type"] in ["INT_INTERFACE_L", "DSP_L"]:
        bgcolor = "#ffaaff"
    if tiledata["type"] in ["INT_INTERFACE_R", "DSP_R"]:
        bgcolor = "#ffaaff"

    if "IO" in tiledata["type"]:
        bgcolor = "#dddddd"

    # Unused - grey
    if tiledata["type"] in ["NULL", "VBRK"] or "BRK" in tiledata["type"]:
        bgcolor = "#aaaaaa"

    return bgcolor


def tile_print_td(f, title, tilename, bgcolor, tiledata, dbstate):
    tilename = tilename.replace("INT_INTERFACE_", "INTF_")
    tilename = tilename.replace("_X", "<br/>X")
    tilename = tilename.replace("B_TERM", "B<br/>TERM")

    print(
        "<td bgcolor=\"%s\" align=\"center\" title=\"%s\"><span style=\"font-size:10px\">"
        % (bgcolor, "\n".join(title)),
        file=f)
    if tiledata["type"].lower() in dbstate.cfgbits:
        print(
            "<a style=\"text-decoration: none; color: black\" href=\"tile_%s.html\">%s</a></span></td>"
            % (tiledata["type"].lower(), tilename.replace("_X", "<br/>X")),
            file=f)
    else:
        print(
            "%s</span></td>" % tilename.replace("_X", "<br/>X").replace(
                "B_TERM", "B<br/>TERM"),
            file=f)


def tile_title(tilename, tiledata, grid_x, grid_y, grid):
    title = [tilename]
    segdata = None

    if "segment" in tiledata:
        # FIXME: BRAM support
        segdata = grid[tilename]['bits'].get('CLB_IO_CLK', None)
        title.append(tiledata["segment"])

    title.append("GRID_POSITION: %d %d" % (grid_x, grid_y))

    if "sites" in tiledata:
        for sitename, sitetype in tiledata["sites"].items():
            title.append("%s site: %s" % (sitetype, sitename))

    if segdata:
        if "baseaddr" in segdata:
            #title.append("Baseaddr: %s %d" % tuple(segdata["baseaddr"]))
            title.append("Baseaddr: %s" % segdata["baseaddr"])
        else:
            print(
                "Warning: no baseaddr in segment %s (via tile %s)." %
                (tiledata["segment"], tilename))

    return title


def mk_tilegrid_page(dbstate, output, grid):
    with out_open("index.html", output) as f:
        print(
            "<html><title>X-Ray %s Database</title><body>" %
            get_setting("XRAY_DATABASE").upper(),
            file=f)
        print(
            "<h3>X-Ray %s Database</h3>" %
            get_setting("XRAY_DATABASE").upper(),
            file=f)

        print(
            "<p><b>Part: %s<br/>ROI TILEGRID: %s<br/>ROI Frames: %s</b></p>" %
            (
                get_setting("XRAY_PART"), get_setting("XRAY_ROI_TILEGRID"),
                get_setting("XRAY_ROI_FRAMES")),
            file=f)

        grid_map, grid_range = place_tiles(grid)

        print("<table border>", file=f)

        for grid_y in range(grid_range[1], grid_range[3] + 1):
            print("<tr>", file=f)

            for grid_x in range(grid_range[0], grid_range[2] + 1):
                tilename = grid_map[(grid_x, grid_y)]
                tiledata = grid[tilename]

                bgcolor = tile_bgcolor(tiledata)
                title = tile_title(tilename, tiledata, grid_x, grid_y, grid)
                tile_print_td(f, title, tilename, bgcolor, tiledata, dbstate)

            print("</tr>", file=f)

        print("</table>", file=f)
        print("</body></html>", file=f)


def get_bit_name(dbstate, frameidx, bitidx, bit_pos, tiletype):
    bit_name = dbstate.cfgbits_r[tiletype][
        bit_pos] if bit_pos in dbstate.cfgbits_r[tiletype] else None

    if bit_name is None and bit_pos in dbstate.routebits[tiletype]:
        bit_name = dbstate.routebits[tiletype][bit_pos]

    if bit_name is None and bit_pos in dbstate.routezbits[tiletype]:
        bit_name = dbstate.routezbits[tiletype][bit_pos]

    if bit_name is None and tiletype in ["clbll_l", "clbll_r", "clblm_l",
                                         "clblm_r", "dsp_l", "dsp_r", "bram_l",
                                         "bram_r"]:
        int_tile_type = "int_" + tiletype[-1]
        bit_int_pos = "%02d_%02d" % (frameidx, bitidx % 64)
        bit_name = dbstate.cfgbits_r[int_tile_type][
            bit_int_pos] if bit_int_pos in dbstate.cfgbits_r[
                int_tile_type] else None

        if bit_name is None and bit_int_pos in dbstate.routebits[int_tile_type]:
            bit_name = dbstate.routebits[int_tile_type][bit_int_pos]

        if bit_name is None and bit_int_pos in dbstate.routezbits[
                int_tile_type]:
            bit_name = dbstate.routezbits[int_tile_type][bit_int_pos]

    if bit_name is not None:
        if len(bit_name) <= 1:
            bit_name = "".join(bit_name)
        else:
            for n in bit_name:
                bit_name = ".".join(n.split(".")[:-1])

    return bit_name


def get_bit_info(dbstate, frameidx, bitidx, tiletype):
    bit_pos = "%02d_%02d" % (frameidx, bitidx)

    bit_name = get_bit_name(dbstate, frameidx, bitidx, bit_pos, tiletype)

    label = None
    title = [bit_pos]
    bgcolor = "#aaaaaa"

    if bit_pos not in dbstate.maskbits[tiletype]:
        label = "&nbsp;"
        bgcolor = "#444444"
        title.append("UNUSED ?")

    if (tiletype in ["hclk_l", "hclk_r"]) and bitidx < 13:
        label = "ECC"
        bgcolor = "#44aa44"

    if bit_name is not None:
        bgcolor = "#ff0000"
        title.append(bit_name)

        if "LUT.INIT" in bit_name:
            bgcolor = "#ffffaa"
            m = re.search(r"(.)LUT.INIT\[(..)\]", bit_name)
            label = m.group(1) + m.group(2)

        m = re.search(r"\.([ABCD])LUT\.([A-Z]+)$", bit_name)
        if m:
            bgcolor = "#ffffaa"
            if m.group(2) == "RAM":
                label = m.group(1) + "LR"
            elif m.group(2) == "SMALL":
                label = m.group(1) + "LS"
            elif m.group(2) == "SRL":
                label = m.group(1) + "SR"
            else:
                bgcolor = "#ff0000"

        m = re.search(r"\.([ABCD])LUT\.([A-Z]+)$", bit_name)
        if m:
            bgcolor = "#ffffaa"
            if m.group(2) == "RAM":
                label = m.group(1) + "LR"
            elif m.group(2) == "SMALL":
                label = m.group(1) + "LS"
            elif m.group(2) == "SRL":
                label = m.group(1) + "SR"
            else:
                bgcolor = "#ff0000"

        m = re.search(r"\.([ABCD]5?)FF\.([A-Z]+(\.A|\.B)?)$", bit_name)
        if m:
            bgcolor = "#aaffaa"
            if m.group(2) == "ZINI":
                label = m.group(1) + "ZI"
            elif m.group(2) == "ZRST":
                label = m.group(1) + "ZR"
            elif m.group(2) == "MUX.A":
                label = m.group(1) + "MA"
            elif m.group(2) == "MUX.B":
                label = m.group(1) + "MB"
            else:
                bgcolor = "#ff0000"

        m = re.search(r"\.([ABCD])LUT.DI1MUX\.", bit_name)
        if m:
            bgcolor = "#ffffaa"
            label = m.group(1) + "DI1"

        m = re.search(r"\.(WA[78])USED$", bit_name)
        if m:
            bgcolor = "#ffffaa"
            label = m.group(1)

        if ".WEMUX." in bit_name:
            bgcolor = "#ffffaa"
            label = "WE"

        m = re.search(r"\.CARRY4\.([A-Z0-9]+)$", bit_name)
        if m:
            bgcolor = "#88cc00"
            label = m.group(1)

        if re.search(r"\.LATCH$", bit_name):
            bgcolor = "#aaffaa"
            label = "LAT"

        if re.search(r"\.FFSYNC$", bit_name):
            bgcolor = "#aaffaa"
            label = "SYN"

        if re.search(r"\.[ABCD]LUT5$", bit_name):
            bgcolor = "#cccc88"
            label = bit_name[-5] + "5"

        if re.search(r"\.(CE|SR)USEDMUX$", bit_name):
            bgcolor = "#ffaa00"
            label = bit_name[-9:-7] + "M"

        if re.search(r"\.CLKINV$", bit_name):
            bgcolor = "#ffaa00"
            label = "CLKI"

        if ".ENABLE_BUFFER." in bit_name:
            bgcolor = "#ffaa00"
            label = "BUF"

        if re.match("^INT_[LR].[SNWE][SNWERL]", bit_name):
            if bit_name[8] == "1":
                bgcolor = "#4466bb"
            elif bit_name[8] == "2":
                bgcolor = "#aa88ff"
            elif bit_name[6:9] in "SS6 SE6 NN6 NW6".split():
                bgcolor = "#7755ff"
            else:
                bgcolor = "#88aaff"
            label = bit_name[6:9]

        if re.match("^INT_[LR].IMUX", bit_name):
            m = re.match("^INT_[LR].IMUX(_L)?(\\d+)", bit_name)
            bgcolor = "#88aaff"
            label = "IM" + m.group(2)

        if re.match("^INT_[LR].BYP_ALT", bit_name):
            bgcolor = "#7755ff"
            label = "BA" + bit_name[13]

        if re.match("^INT_[LR].FAN_ALT", bit_name):
            bgcolor = "#4466bb"
            label = "FA" + bit_name[13]

        if re.match("^INT_[LR].CLK", bit_name):
            bgcolor = "#4466bb"
            label = "CLK"

        if re.match("^INT_[LR].CTRL", bit_name):
            bgcolor = "#7755ff"
            label = "CTRL"

        if re.match("^INT_[LR].GFAN", bit_name):
            bgcolor = "#7755ff"
            label = "GFAN"

        if re.match("^INT_[LR].LVB", bit_name):
            bgcolor = "#88aaff"
            label = "LVB"

        if re.match("^INT_[LR].LV", bit_name):
            bgcolor = "#88aaff"
            label = "LV"

        if re.match("^INT_[LR].LH", bit_name):
            bgcolor = "#4466bb"
            label = "LH"

        if re.match("^CLBL[LM]_[LR].SLICE[LM]_X[01].[ABCD]5?FFMUX", bit_name):
            bgcolor = "#88aaff"
            label = "DMX"

        if re.match("^CLBL[LM]_[LR].SLICE[LM]_X[01].[ABCD]OUTMUX", bit_name):
            bgcolor = "#aa88ff"
            label = "OMX"

        if re.match("^CLBL[LM]_[LR].SLICE[LM]_X[01].PRECYINIT", bit_name):
            bgcolor = "#88aaff"
            label = "CYI"

        if re.match("^HCLK_[LR]", bit_name) and "_B_BOT" in bit_name:
            bgcolor = "#4466bb"
            label = "BOT"

        if re.match("^HCLK_[LR]", bit_name) and "_B_TOP" in bit_name:
            bgcolor = "#7755ff"
            label = "TOP"

        if re.match("^DSP_[LR].DSP_[01].(PATTERN|MASK)", bit_name):
            bgcolor = "#ffffaa"
            label = bit_name[12] + bit_name[10]

    return bit_pos, label, title, bgcolor


def gen_table(dbstate, tiletype, f):
    print(
        """
<script><!--
var grp2bits = { };
var bit2grp = { }
var highlight_bits = [ ];
var highlight_cache = { };

function ome(bit) {
    // console.log("ome: " + bit);
    if (bit in bit2grp) {
        grp = bit2grp[bit];
        for (i in grp2bits[grp]) {
            b = grp2bits[grp][i];
            // console.log("  -> " + b);
            if (!(b in highlight_cache)) {
                el = document.getElementById("bit" + b);
                highlight_cache[b] = el.bgColor;
                el.bgColor = "#ffffff";
                highlight_bits.push(b);
            }
        }
    }
}

function oml() {
    // console.log("oml");
    for (i in highlight_bits) {
        b = highlight_bits[i];
        el = document.getElementById("bit" + b);
        el.bgColor = highlight_cache[b];
        delete highlight_cache[b];
        el.style.fontWeight = "normal";
    }
    highlight_bits.length = 0;
}
//--></script>
    """,
        file=f)

    num_frames = 36
    unused_bits = 0
    unknown_bits = 0
    known_bits = 0
    hideframes = set()

    if tiletype in ["int_l", "int_r", "hclk_l", "hclk_r"]:
        num_frames = 26

    if tiletype in ["bram_l", "bram_r", "dsp_l", "dsp_r"]:
        for i in range(5, 25):
            hideframes.add(i)
        num_frames = 28

    height = 2
    if tiletype in ["hclk_l", "hclk_r"]:
        height = 1
    if tiletype in ["dsp_l", "dsp_r", "bram_l", "bram_r"]:
        height = 10

    if height > 2:
        print("<table><tr><td>", file=f)

    def print_table_header():
        print("<table border>", file=f)
        print("<tr>", file=f)
        print("<th width=\"30\"></th>", file=f)
        for frameidx in range(num_frames):
            if frameidx in hideframes:
                continue
            print(
                "<th width=\"30\"><span style=\"font-size:10px\">%d</span></th>"
                % frameidx,
                file=f)
        print("</tr>", file=f)

    print_table_header()

    for bitidx in range(32 * height - 1, -1, -1):
        print("<tr>", file=f)
        print(
            "<th align=\"right\"><span style=\"font-size:10px\">%d</span></th>"
            % bitidx,
            file=f)
        for frameidx in range(num_frames):
            if frameidx in hideframes:
                continue

            bit_pos, label, title, bgcolor = get_bit_info(
                dbstate, frameidx, bitidx, tiletype)

            if label is None:
                label = "&nbsp;"

            if label == "INT":
                onclick = " onmousedown=\"location.href = 'tile_int_%s.html#b%s'\"" % (
                    tiletype[-1], bit_pos)
            else:
                onclick = " onmousedown=\"location.href = '#b%s'\"" % bit_pos

            if bgcolor == "#aaaaaa":
                unknown_bits += 1
            elif bgcolor == "#444444":
                unused_bits += 1
            else:
                known_bits += 1

            print(
                "<td id=\"bit%s\" onmouseenter=\"ome('%s');\" onmouseleave=\"oml();\" bgcolor=\"%s\" align=\"center\" title=\"%s\"%s><span style=\"font-size:10px\">%s</span></td>"
                %
                (bit_pos, bit_pos, bgcolor, "\n".join(title), onclick, label),
                file=f)
        print("</tr>", file=f)

        if bitidx > 0 and bitidx % 80 == 0:
            print("</table></td><td>", file=f)
            print_table_header()

    print("</table>", file=f)

    if height > 2:
        print("</td></tr></table>", file=f)

    print(
        "  unused: %d, unknown: %d, known: %d, total: %d, percentage: %.2f%% (%.2f%%)"
        % (
            unused_bits, unknown_bits, known_bits,
            unused_bits + unknown_bits + known_bits, 100 * known_bits /
            (unknown_bits + unused_bits + known_bits), 100 *
            (known_bits + unused_bits) /
            (unknown_bits + unused_bits + known_bits)))


def mk_segment_pages(dbstate, output, tweaks):
    for tiletype in sorted(dbstate.cfgbits.keys()):
        with out_open("tile_%s.html" % tiletype, output) as f:
            print(
                "<html><title>X-Ray %s Database: %s</title><body>" %
                (get_setting("XRAY_DATABASE").upper(), tiletype.upper()),
                file=f)
            print(
                "<h3><a href=\"index.html\">X-Ray %s Database</a>: %s Segment</h3>"
                % (get_setting("XRAY_DATABASE").upper(), tiletype.upper()),
                file=f)

            gen_table(dbstate, tiletype, f)

            print("<div>", file=f)

            bits_by_prefix = dict()

            for bit_name, bits_pos in dbstate.cfgbits[tiletype].items():
                prefix = ".".join(bit_name.split(".")[0:-1])

                if prefix not in bits_by_prefix:
                    bits_by_prefix[prefix] = set()

                for bit_pos in bits_pos:
                    bits_by_prefix[prefix].add((bit_name, bit_pos))

            for prefix, bits in sorted(bits_by_prefix.items()):
                for bit_name, bit_pos in sorted(bits):
                    print("<a id=\"b%s\"/>" % bit_pos, file=f)

                print("<p/>", file=f)
                print("<h4>%s</h4>" % prefix, file=f)
                print("<table cellspacing=0>", file=f)
                print(
                    "<tr><th width=\"400\" align=\"left\">Bit Name</th><th>Position</th></tr>",
                    file=f)

                trstyle = ""
                for bit_name, bit_pos in sorted(bits):
                    trstyle = " bgcolor=\"#dddddd\"" if trstyle == "" else ""
                    print(
                        "<tr%s><td>%s</td><td>%s</td></tr>" %
                        (trstyle, bit_name, bit_pos),
                        file=f)

                print("</table>", file=f)

            ruf = UnionFind()
            routebits_routezbits = list(dbstate.routebits[tiletype].items())
            routebits_routezbits += list(dbstate.routezbits[tiletype].items())

            for bit, pips in routebits_routezbits:
                for pip in pips:
                    grp = ".".join(pip.split('.')[:-1])
                    ruf.union(grp, bit)

            rgroups = dict()
            rgroup_names = dict()

            for bit, pips in routebits_routezbits:
                for pip in pips:
                    grp_name = ".".join(pip.split('.')[:-1])
                    grp = ruf.find(grp_name)
                    if grp not in rgroup_names:
                        rgroup_names[grp] = set()
                    rgroup_names[grp].add(grp_name)
                    if grp not in rgroups:
                        rgroups[grp] = dict()
                    if pip not in rgroups[grp]:
                        rgroups[grp][pip] = set()
                    rgroups[grp][pip].add(bit)

            shared_bits = dict()
            for bit, pips in routebits_routezbits:
                for pip in pips:
                    grp_name = ".".join(pip.split('.')[:-1])
                    if bit not in shared_bits:
                        shared_bits[bit] = set()
                    shared_bits[bit].add(grp_name)

            rgroups_with_title = list()

            for grp, gdata in sorted(rgroups.items()):
                title = "PIPs driving " + ", ".join(sorted(rgroup_names[grp]))
                rgroups_with_title.append((title, grp, gdata))

            for title, grp, gdata in sorted(rgroups_with_title):
                grp_bits = set()
                for pip, bits in gdata.items():
                    grp_bits |= bits

                def bit_key(b):
                    if tiletype in ["hclk_l", "hclk_r"]:
                        if b in tweaks.hclk_left_bits:
                            return "a" + b
                        if b in tweaks.hclk_right_bits:
                            return "c" + b
                    if tiletype in ["clblm_l", "clblm_r", "clbll_l", "clbll_r",
                                    "int_l", "int_r"]:
                        if b in tweaks.clb_left_bits:
                            return "a" + b
                        if b in tweaks.clb_right_bits:
                            return "c" + b
                    return "b" + b

                grp_bits = sorted(grp_bits, key=bit_key)

                for bit in grp_bits:
                    print("<a id=\"b%s\"/>" % bit, file=f)

                print("<script><!--", file=f)
                print(
                    "grp2bits['%s'] = ['%s'];" %
                    (grp_bits[0], "', '".join(grp_bits)),
                    file=f)
                for bit in grp_bits:
                    print("bit2grp['%s'] = '%s';" % (bit, grp_bits[0]), file=f)
                print("//--></script>", file=f)

                print("<p/>", file=f)
                print("<h4>%s</h4>" % title, file=f)
                print("<table cellspacing=0>", file=f)
                print("<tr><th width=\"400\" align=\"left\">PIP</th>", file=f)

                for bit in grp_bits:
                    print("<th>&nbsp;%s&nbsp;</th>" % bit, file=f)
                print("</tr>", file=f)

                lines = list()
                for pip, bits in sorted(gdata.items()):
                    line = " --><td>%s</td>" % (pip)
                    for bit in grp_bits:
                        c = "-"
                        if bit in dbstate.routebits[
                                tiletype] and pip in dbstate.routebits[
                                    tiletype][bit]:
                            c = "1"
                        if bit in dbstate.routezbits[
                                tiletype] and pip in dbstate.routezbits[
                                    tiletype][bit]:
                            c = "0"
                        line = "%s%s<td align=\"center\">%s</td>" % (
                            c, line, c)
                    lines.append(line)

                trstyle = ""
                for line in sorted(lines):
                    trstyle = " bgcolor=\"#dddddd\"" if trstyle == "" else ""
                    print("<tr%s><!-- %s</tr>" % (trstyle, line), file=f)

                print("</table>", file=f)

                first_note = True
                for bit in grp_bits:
                    if len(shared_bits[bit]) > 1:
                        if first_note:
                            print("<p><b>Note(s):</b><br/>", file=f)
                        print(
                            "Warning: Groups sharing bit %s: %s." %
                            (bit, ", ".join(sorted(shared_bits[bit]))))
                        print(
                            "Groups sharing bit <b>%s</b>: %s.<br/>" %
                            (bit, ", ".join(sorted(shared_bits[bit]))),
                            file=f)
                        first_note = False
                if not first_note:
                    print("</p>", file=f)

            if len(dbstate.ppips[tiletype]) > 0:
                print("<h4>Pseudo PIPs</h4>", file=f)
                print("<table cellspacing=0>", file=f)
                print(
                    "<tr><th width=\"500\" align=\"left\">PIP</th><th>Type</th></tr>",
                    file=f)
                trstyle = ""
                for typ, tag in sorted(
                    [(b, a) for a, b in dbstate.ppips[tiletype].items()]):
                    trstyle = " bgcolor=\"#dddddd\"" if trstyle == "" else ""
                    print(
                        "<tr%s><td>%s</td><td>%s</td></tr>" %
                        (trstyle, tag, typ),
                        file=f)
                print("</table>", file=f)

            print("</div>", file=f)
            print("</body></html>", file=f)


def run(settings, output, verbose=False, allow_fake=False):
    global get_setting

    get_setting = mk_get_setting(settings)

    if output is None:
        output = os.path.join(
            os.path.curdir, 'html', get_setting('XRAY_DATABASE'))

    db_dir = os.path.join(
        get_setting("XRAY_DATABASE_DIR"), get_setting("XRAY_DATABASE"))

    # Load tweaks
    tweaks = Tweaks()
    clb_bitgroups_db, hclk_bitgroups_db = init_bitdb()
    tweaks.hclk_left_bits, tweaks.hclk_right_bits = init_hclk_bits(
        hclk_bitgroups_db)
    tweaks.clb_left_bits, tweaks.clb_right_bits = init_clb_bits(
        clb_bitgroups_db)

    # Load source data
    dbstate = DBState()
    fabric = get_fabric_for_part(db_dir, get_setting("XRAY_PART"))
    grid = load_tilegrid(
        db_dir, fabric, verbose=verbose, allow_fake=allow_fake)
    db_reads(dbstate, db_dir)

    # Create pages
    mk_tilegrid_page(dbstate, output, grid)
    mk_segment_pages(dbstate, output, tweaks)


def main():
    import argparse

    parser = argparse.ArgumentParser(
        description="Generate a pretty HTML version of the documentation.")
    parser.add_argument('--verbose', action='store_true')
    parser.add_argument(
        '--output',
        default=None,
        help='Put the generated files in this directory (default current dir).'
    )
    parser.add_argument(
        '--settings',
        default=None,
        help='Read the settings from file (default to environment).',
    )
    parser.add_argument(
        '--allow-fake',
        default=False,
        action='store_true',
        help="Continue even if tilegrid.json isn't found.",
    )

    args = parser.parse_args()
    run(
        settings=args.settings,
        output=args.output,
        verbose=args.verbose,
        allow_fake=args.allow_fake,
    )


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

from prjxray.segmaker import Segmaker

segmk = Segmaker("design.bits")

print("Loading tags")
'''
name,loc,ce,r
clb_FDRE,SLICE_X12Y100,1,0
clb_FDRE,SLICE_X13Y100,1,1
clb_FDRE,SLICE_X14Y100,1,1
'''
f = open('params.csv', 'r')
f.readline()
for l in f:
    name, site, ce, r = l.split(',')
    ce = int(ce)
    r = int(r)

    # Theory: default position are the force positions
    # parameter FORCE_CE1=0;
    # parameter nFORCE_R0=1;
    # .CE(din[0] | FORCE_CE1),
    # .R(din[1] & nFORCE_R0),
    segmk.add_site_tag(site, "CEUSEDMUX", ce ^ 1)
    segmk.add_site_tag(site, "SRUSEDMUX", r)
segmk.compile()
segmk.write()
